#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

import Badmunda
from Badmunda import raw, types, utils

class SendScreenshotNotification:
    async def send_screenshot_notification(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        reply_parameters: "types.ReplyParameters" = None
    ) -> "types.Message":
        """Notify the other user in a private chat that a screenshot of the chat was taken.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            reply_parameters (:obj:`~Badmunda.types.ReplyParameters`, *optional*):
                Describes reply parameters for the message that is being sent.

        Returns:
            :obj:`~Badmunda.types.Message`: On success, the sent service message is returned.
        """

        r = await self.invoke(
            raw.functions.messages.SendScreenshotNotification(
                peer=await self.resolve_peer(chat_id),
                reply_to=await utils.get_reply_to(self, reply_parameters) if reply_parameters else raw.types.InputReplyToMessage(reply_to_msg_id=0),
                random_id=self.rnd_id()
            )
        )

        messages = await utils.parse_messages(client=self, messages=r)

        return messages[0] if messages else None



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import Badmunda
from Badmunda import raw, enums


class SearchGlobalCount:
    async def search_global_count(
        self: "Badmunda.Client",
        query: str = "",
        filter: "enums.MessagesFilter" = enums.MessagesFilter.EMPTY,
        channels_only: Optional[bool] = None,
        groups_only: Optional[bool] = None,
        users_only: Optional[bool] = None,
    ) -> int:
        """Get the count of messages resulting from a global search.

        If you want to get the actual messages, see :meth:`~Badmunda.Client.search_global`.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            query (``str``, *optional*):
                Text query string.
                Use "@" to search for mentions.

            filter (:obj:`~Badmunda.enums.MessagesFilter`, *optional*):
                Pass a filter in order to search for specific kind of messages only:

            channels_only (``bool``, *optional*):
                Pass True to search only in channels.

            groups_only (``bool``, *optional*):
                Pass True to search only in groups.

            users_only (``bool``, *optional*):
                Pass True to search only in users.

        Returns:
            ``int``: On success, the messages count is returned.
        """
        r = await self.invoke(
            raw.functions.messages.SearchGlobal(
                q=query,
                filter=filter.value(),
                min_date=0,
                max_date=0,
                offset_rate=0,
                offset_peer=raw.types.InputPeerEmpty(),
                offset_id=0,
                broadcasts_only=channels_only,
                groups_only=groups_only,
                users_only=users_only,
                limit=1
            )
        )

        if hasattr(r, "count"):
            return r.count
        else:
            return len(r.messages)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import List, Union

import Badmunda
from Badmunda import raw, types, utils


class ForwardMediaGroup:
    async def forward_media_group(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        from_chat_id: Union[int, str],
        message_id: int,
        message_thread_id: int = None,
        disable_notification: bool = None,
        schedule_date: datetime = None,
        hide_sender_name: bool = None,
        hide_captions: bool = None,
        protect_content: bool = None,
        allow_paid_broadcast: bool = None,
        reply_parameters: "types.ReplyParameters" = None,
        paid_message_star_count: int = None
    ) -> List["types.Message"]:
        """Forward a media group by providing one of the message ids.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            from_chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the source chat where the original message was sent.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            message_id (``int``):
                Message identifier in the chat specified in *from_chat_id*.

            message_thread_id (``int``, *optional*):
                Unique identifier of a message thread to which the message belongs.
                For supergroups only.

            disable_notification (``bool``, *optional*):
                Sends the message silently.
                Users will receive a notification with no sound.

            schedule_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the message will be automatically sent.

            hide_sender_name (``bool``, *optional*):
                If True, the original author of the message will not be shown.

            hide_captions (``bool``, *optional*):
                If True, the original media captions will be removed.

            protect_content (``bool``, *optional*):
                Protects the contents of the sent message from forwarding and saving.

            allow_paid_broadcast (``bool``, *optional*):
                If True, you will be allowed to send up to 1000 messages per second.
                Ignoring broadcasting limits for a fee of 0.1 Telegram Stars per message.
                The relevant Stars will be withdrawn from the bot's balance.
                For bots only.

            reply_parameters (:obj:`~Badmunda.types.ReplyParameters`, *optional*):
                Describes reply parameters for the message that is being sent.

            paid_message_star_count (``int``, *optional*):
                The number of Telegram Stars the user agreed to pay to send the messages.

        Returns:
            List of :obj:`~Badmunda.types.Message`: On success, a list of forwarded messages is returned.

        Example:
            .. code-block:: python

                # Forward a media group
                await app.forward_media_group(to_chat, from_chat, 123)
        """
        message_ids = [i.id for i in await self.get_media_group(from_chat_id, message_id)]

        r = await self.invoke(
            raw.functions.messages.ForwardMessages(
                to_peer=await self.resolve_peer(chat_id),
                from_peer=await self.resolve_peer(from_chat_id),
                id=message_ids,
                silent=disable_notification or None,
                random_id=[self.rnd_id() for _ in message_ids],
                schedule_date=utils.datetime_to_timestamp(schedule_date),
                drop_author=hide_sender_name,
                drop_media_captions=hide_captions,
                noforwards=protect_content,
                allow_paid_floodskip=allow_paid_broadcast,
                reply_to=await utils.get_reply_to(
                    client=self,
                    reply_parameters=reply_parameters,
                    message_thread_id=message_thread_id
                ),
                allow_paid_stars=paid_message_star_count
            )
        )

        return await utils.parse_messages(client=self, messages=r)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import typing
from datetime import datetime
from enum import Enum
from json import dumps

import Badmunda


class Object:
    def __init__(self, client: "Badmunda.Client" = None):
        self._client = client

    def bind(self, client: "Badmunda.Client"):
        """Bind a Client instance to this and to all nested Badmunda objects.

        Parameters:
            client (:obj:`~Badmunda.types.Client`):
                The Client instance to bind this object with. Useful to re-enable bound methods after serializing and
                deserializing Badmunda objects with ``repr`` and ``eval``.
        """
        self._client = client

        for i in self.__dict__:
            o = getattr(self, i)

            if isinstance(o, Object):
                o.bind(client)

    @staticmethod
    def default(obj: "Object"):
        if isinstance(obj, bytes):
            return repr(obj)

        # https://t.me/Badmundachat/167281
        # Instead of re.Match, which breaks for python <=3.6
        if isinstance(obj, typing.Match):
            return repr(obj)

        if isinstance(obj, Enum):
            return str(obj)

        if isinstance(obj, datetime):
            return str(obj)

        attributes_to_hide = [
            "raw"
        ]

        filtered_attributes = {
            attr: ("*" * 9 if attr == "phone_number" else getattr(obj, attr))
            for attr in filter(
                lambda x: not x.startswith("_") and x not in attributes_to_hide,
                obj.__dict__,
            )
            if getattr(obj, attr) is not None
        }

        return {
            "_": obj.__class__.__name__,
            **filtered_attributes
        }

    def __str__(self) -> str:
        return dumps(self, indent=4, default=Object.default, ensure_ascii=False)

    def __repr__(self) -> str:
        return "Badmunda.types.{}({})".format(
            self.__class__.__name__,
            ", ".join(
                f"{attr}={repr(getattr(self, attr))}"
                for attr in filter(lambda x: not x.startswith("_"), self.__dict__)
                if getattr(self, attr) is not None
            )
        )

    def __eq__(self, other: "Object") -> bool:
        for attr in self.__dict__:
            try:
                if attr.startswith("_"):
                    continue

                if getattr(self, attr) != getattr(other, attr):
                    return False
            except AttributeError:
                return False

        return True

    def __setstate__(self, state):
        for attr in state:
            obj = state[attr]

            # Maybe a better alternative would be https://docs.python.org/3/library/inspect.html#inspect.signature
            if isinstance(obj, tuple) and len(obj) == 2 and obj[0] == "dt":
                state[attr] = datetime.fromtimestamp(obj[1])

        self.__dict__ = state

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("_client", None)

        for attr in state:
            obj = state[attr]

            if isinstance(obj, datetime):
                state[attr] = ("dt", obj.timestamp())

        return state



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
from datetime import datetime
from typing import BinaryIO, Callable, List, Optional, Union

import Badmunda
from Badmunda import StopTransmission, enums, raw, types, utils
from Badmunda.errors import FilePartMissing
from Badmunda.file_id import FileType

log = logging.getLogger(__name__)

class SendAudio:
    async def send_audio(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        audio: Union[str, BinaryIO],
        caption: str = "",
        parse_mode: Optional["enums.ParseMode"] = None,
        caption_entities: List["types.MessageEntity"] = None,
        duration: int = 0,
        performer: str = None,
        title: str = None,
        thumb: Union[str, BinaryIO] = None,
        file_name: str = None,
        disable_notification: bool = None,
        message_thread_id: int = None,
        direct_messages_chat_topic_id: int = None,
        effect_id: int = None,
        reply_parameters: "types.ReplyParameters" = None,
        schedule_date: datetime = None,
        protect_content: bool = None,
        business_connection_id: str = None,
        allow_paid_broadcast: bool = None,
        paid_message_star_count: int = None,
        reply_markup: Union[
            "types.InlineKeyboardMarkup",
            "types.ReplyKeyboardMarkup",
            "types.ReplyKeyboardRemove",
            "types.ForceReply"
        ] = None,
        progress: Callable = None,
        progress_args: tuple = (),

        reply_to_message_id: int = None,
        reply_to_chat_id: Union[int, str] = None,
        reply_to_story_id: int = None,
        quote_text: str = None,
        quote_entities: List["types.MessageEntity"] = None,
        quote_offset: int = None,
    ) -> Optional["types.Message"]:
        """Send audio files.

        For sending voice messages, use the :meth:`~Badmunda.Client.send_voice` method instead.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            audio (``str`` | ``BinaryIO``):
                Audio file to send.
                Pass a file_id as string to send an audio file that exists on the Telegram servers,
                pass an HTTP URL as a string for Telegram to get an audio file from the Internet,
                pass a file path as string to upload a new audio file that exists on your local machine, or
                pass a binary file-like object with its attribute ".name" set for in-memory uploads.

            caption (``str``, *optional*):
                Audio caption, 0-1024 characters.

            parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            caption_entities (List of :obj:`~Badmunda.types.MessageEntity`):
                List of special entities that appear in the caption, which can be specified instead of *parse_mode*.

            duration (``int``, *optional*):
                Duration of the audio in seconds.

            performer (``str``, *optional*):
                Performer.

            title (``str``, *optional*):
                Track name.

            thumb (``str`` | ``BinaryIO``, *optional*):
                Thumbnail of the music file album cover.
                The thumbnail should be in JPEG format and less than 200 KB in size.
                A thumbnail's width and height should not exceed 320 pixels.
                Thumbnails can't be reused and can be only uploaded as a new file.

            file_name (``str``, *optional*):
                File name of the audio sent.
                Defaults to file's path basename.

            disable_notification (``bool``, *optional*):
                Sends the message silently.
                Users will receive a notification with no sound.

            message_thread_id (``int``, *optional*):
                Unique identifier for the target message thread (topic) of the forum.
                For forums only.

            direct_messages_chat_topic_id (``int``, *optional*):
                Unique identifier of the topic in a channel direct messages chat administered by the current user.
                For directs only only.

            effect_id (``int``, *optional*):
                Unique identifier of the message effect.
                For private chats only.

            reply_parameters (:obj:`~Badmunda.types.ReplyParameters`, *optional*):
                Describes reply parameters for the message that is being sent.

            schedule_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the message will be automatically sent.

            protect_content (``bool``, *optional*):
                Protects the contents of the sent message from forwarding and saving.

            business_connection_id (``str``, *optional*):
                Unique identifier of the business connection on behalf of which the message will be sent.

            allow_paid_broadcast (``bool``, *optional*):
                If True, you will be allowed to send up to 1000 messages per second.
                Ignoring broadcasting limits for a fee of 0.1 Telegram Stars per message.
                The relevant Stars will be withdrawn from the bot's balance.
                For bots only.

            paid_message_star_count (``int``, *optional*):
                The number of Telegram Stars the user agreed to pay to send the messages.

            reply_markup (:obj:`~Badmunda.types.InlineKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardRemove` | :obj:`~Badmunda.types.ForceReply`, *optional*):
                Additional interface options. An object for an inline keyboard, custom reply keyboard,
                instructions to remove reply keyboard or to force a reply from the user.

            progress (``Callable``, *optional*):
                Pass a callback function to view the file transmission progress.
                The function must take *(current, total)* as positional arguments (look at Other Parameters below for a
                detailed description) and will be called back each time a new file chunk has been successfully
                transmitted.

            progress_args (``tuple``, *optional*):
                Extra custom arguments for the progress callback function.
                You can pass anything you need to be available in the progress callback scope; for example, a Message
                object or a Client instance in order to edit the message with the updated progress status.

        Other Parameters:
            current (``int``):
                The amount of bytes transmitted so far.

            total (``int``):
                The total size of the file.

            *args (``tuple``, *optional*):
                Extra custom arguments as defined in the ``progress_args`` parameter.
                You can either keep ``*args`` or add every single extra argument in your function signature.

        Returns:
            :obj:`~Badmunda.types.Message` | ``None``: On success, the sent audio message is returned, otherwise, in
            case the upload is deliberately stopped with :meth:`~Badmunda.Client.stop_transmission`, None is returned.

        Example:
            .. code-block:: python

                # Send audio file by uploading from file
                await app.send_audio("me", "audio.mp3")

                # Add caption to the audio
                await app.send_audio("me", "audio.mp3", caption="audio caption")

                # Set audio metadata
                await app.send_audio(
                    "me", "audio.mp3",
                    title="Title", performer="Performer", duration=234)

                # Keep track of the progress while uploading
                async def progress(current, total):
                    print(f"{current * 100 / total:.1f}%")

                await app.send_audio("me", "audio.mp3", progress=progress)
        """
        if any(
            (
                reply_to_message_id is not None,
                reply_to_chat_id is not None,
                reply_to_story_id is not None,
                quote_text is not None,
                quote_entities is not None,
                quote_offset is not None,
            )
        ):
            if reply_to_message_id is not None:
                log.warning(
                    "`reply_to_message_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if reply_to_chat_id is not None:
                log.warning(
                    "`reply_to_chat_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if reply_to_story_id is not None:
                log.warning(
                    "`reply_to_story_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_text is not None:
                log.warning(
                    "`quote_text` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_entities is not None:
                log.warning(
                    "`quote_entities` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_offset is not None:
                log.warning(
                    "`quote_offset` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            reply_parameters = types.ReplyParameters(
                message_id=reply_to_message_id,
                chat_id=reply_to_chat_id,
                story_id=reply_to_story_id,
                quote=quote_text,
                quote_parse_mode=parse_mode,
                quote_entities=quote_entities,
                quote_position=quote_offset
            )

        file = None

        try:
            if isinstance(audio, str):
                if os.path.isfile(audio):
                    mime_type = self.guess_mime_type(audio) or "audio/mpeg"
                    if mime_type == "audio/ogg":
                        mime_type = "audio/opus"
                    thumb = await self.save_file(thumb)
                    file = await self.save_file(audio, progress=progress, progress_args=progress_args)
                    media = raw.types.InputMediaUploadedDocument(
                        mime_type=mime_type,
                        file=file,
                        thumb=thumb,
                        attributes=[
                            raw.types.DocumentAttributeAudio(
                                duration=duration,
                                performer=performer,
                                title=title
                            ),
                            raw.types.DocumentAttributeFilename(file_name=file_name or os.path.basename(audio))
                        ]
                    )
                elif re.match("^https?://", audio):
                    media = raw.types.InputMediaDocumentExternal(
                        url=audio
                    )
                else:
                    media = utils.get_input_media_from_file_id(audio, FileType.AUDIO)
            else:
                mime_type = self.guess_mime_type(file_name or audio.name) or "audio/mpeg"
                if mime_type == "audio/ogg":
                    mime_type = "audio/opus"
                thumb = await self.save_file(thumb)
                file = await self.save_file(audio, progress=progress, progress_args=progress_args)
                media = raw.types.InputMediaUploadedDocument(
                    mime_type=mime_type,
                    file=file,
                    thumb=thumb,
                    attributes=[
                        raw.types.DocumentAttributeAudio(
                            duration=duration,
                            performer=performer,
                            title=title
                        ),
                        raw.types.DocumentAttributeFilename(file_name=file_name or audio.name)
                    ]
                )

            while True:
                try:
                    peer = await self.resolve_peer(chat_id)
                    r = await self.invoke(
                        raw.functions.messages.SendMedia(
                            peer=peer,
                            media=media,
                            silent=disable_notification or None,
                            reply_to=await utils.get_reply_to(
                                self,
                                reply_parameters,
                                message_thread_id,
                                direct_messages_chat_topic_id
                            ),
                            random_id=self.rnd_id(),
                            schedule_date=utils.datetime_to_timestamp(schedule_date),
                            noforwards=protect_content,
                            allow_paid_floodskip=allow_paid_broadcast,
                            reply_markup=await reply_markup.write(self) if reply_markup else None,
                            effect=effect_id,
                            allow_paid_stars=paid_message_star_count,
                            **await utils.parse_text_entities(self, caption, parse_mode, caption_entities)
                        ),
                        business_connection_id=business_connection_id
                    )
                except FilePartMissing as e:
                    await self.save_file(audio, file_id=file.id, file_part=e.value)
                else:
                    messages = await utils.parse_messages(client=self, messages=r)

                    return messages[0] if messages else None
        except StopTransmission:
            return None



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from .input_checklist import InputChecklist
from .input_contact_message_content import InputContactMessageContent
from .input_credentials import InputCredentials
from .input_credentials_apple_pay import InputCredentialsApplePay
from .input_credentials_google_pay import InputCredentialsGooglePay
from .input_credentials_new import InputCredentialsNew
from .input_credentials_saved import InputCredentialsSaved
from .input_invoice import InputInvoice
from .input_invoice_message import InputInvoiceMessage
from .input_invoice_message_content import InputInvoiceMessageContent
from .input_invoice_name import InputInvoiceName
from .input_location_message_content import InputLocationMessageContent
from .input_media import InputMedia
from .input_media_animation import InputMediaAnimation
from .input_media_audio import InputMediaAudio
from .input_media_document import InputMediaDocument
from .input_media_photo import InputMediaPhoto
from .input_media_video import InputMediaVideo
from .input_message_content import InputMessageContent
from .input_phone_contact import InputPhoneContact
from .input_privacy_rule import InputPrivacyRule
from .input_privacy_rule_allow_all import InputPrivacyRuleAllowAll
from .input_privacy_rule_allow_bots import InputPrivacyRuleAllowBots
from .input_privacy_rule_allow_chats import InputPrivacyRuleAllowChats
from .input_privacy_rule_allow_close_friends import InputPrivacyRuleAllowCloseFriends
from .input_privacy_rule_allow_contacts import InputPrivacyRuleAllowContacts
from .input_privacy_rule_allow_premium import InputPrivacyRuleAllowPremium
from .input_privacy_rule_allow_users import InputPrivacyRuleAllowUsers
from .input_privacy_rule_disallow_all import InputPrivacyRuleDisallowAll
from .input_privacy_rule_disallow_bots import InputPrivacyRuleDisallowBots
from .input_privacy_rule_disallow_chats import InputPrivacyRuleDisallowChats
from .input_privacy_rule_disallow_contacts import InputPrivacyRuleDisallowContacts
from .input_privacy_rule_disallow_users import InputPrivacyRuleDisallowUsers
from .input_text_message_content import InputTextMessageContent
from .input_venue_message_content import InputVenueMessageContent

__all__ = [
    "InputChecklist",
    "InputContactMessageContent",
    "InputCredentials",
    "InputCredentialsApplePay",
    "InputCredentialsGooglePay",
    "InputCredentialsNew",
    "InputCredentialsSaved",
    "InputInvoice",
    "InputInvoiceMessage",
    "InputInvoiceMessageContent",
    "InputInvoiceName",
    "InputLocationMessageContent",
    "InputMedia",
    "InputMediaAnimation",
    "InputMediaAudio",
    "InputMediaDocument",
    "InputMediaPhoto",
    "InputMediaVideo",
    "InputMessageContent",
    "InputPhoneContact",
    "InputPrivacyRule",
    "InputPrivacyRuleAllowAll",
    "InputPrivacyRuleAllowBots",
    "InputPrivacyRuleAllowChats",
    "InputPrivacyRuleAllowCloseFriends",
    "InputPrivacyRuleAllowContacts",
    "InputPrivacyRuleAllowPremium",
    "InputPrivacyRuleAllowUsers",
    "InputPrivacyRuleDisallowAll",
    "InputPrivacyRuleDisallowBots",
    "InputPrivacyRuleDisallowChats",
    "InputPrivacyRuleDisallowContacts",
    "InputPrivacyRuleDisallowUsers",
    "InputTextMessageContent",
    "InputVenueMessageContent"
]



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from Badmunda import raw
from ..object import Object


class BusinessWeeklyOpen(Object):
    """Business weekly open hours.

    Parameters:
        start_minute (``int``):
            Start minute of the working day.

        end_minute (``int``):
            End minute of the working day.
    """

    def __init__(
        self,
        *,
        start_minute: int,
        end_minute: int,

    ):
        self.start_minute = start_minute
        self.end_minute = end_minute

    @staticmethod
    def _parse(weekly_open: "raw.types.BusinessWeeklyOpen" = None) -> "BusinessWeeklyOpen":
        return BusinessWeeklyOpen(
            start_minute=weekly_open.start_minute,
            end_minute=weekly_open.end_minute,
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Optional

import Badmunda
from Badmunda import raw, utils


class RequestCallbackAnswer:
    async def request_callback_answer(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        message_id: int,
        callback_data: Union[str, bytes],
        password: Optional[str] = None,
        timeout: int = 10
    ):
        """Request a callback answer from bots.
        This is the equivalent of clicking an inline button containing callback data.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            message_id (``int``):
                The message id the inline keyboard is attached on.

            callback_data (``str`` | ``bytes``):
                Callback data associated with the inline button you want to get the answer from.

            password (``str``, *optional*):
                When clicking certain buttons (such as BotFather's confirmation button to transfer ownership), if your account has 2FA enabled, you need to provide your account's password.
                The 2-step verification password for the current user. Only applicable, if the :obj:`~Badmunda.types.InlineKeyboardButton` contains ``callback_data_with_password``.

            timeout (``int``, *optional*):
                Timeout in seconds.

        Returns:
            The answer containing info useful for clients to display a notification at the top of the chat screen
            or as an alert.

        Raises:
            TimeoutError: In case the bot fails to answer within 10 seconds.
            ValueError: In case of invalid arguments.
            RPCError: In case of Telegram RPC error.

        Example:
            .. code-block:: python

                await app.request_callback_answer(chat_id, message_id, "callback_data")
        """

        # Telegram only wants bytes, but we are allowed to pass strings too.
        data = bytes(callback_data, "utf-8") if isinstance(callback_data, str) else callback_data

        if password:
            r = await self.invoke(
                raw.functions.account.GetPassword()
            )
            password = utils.compute_password_check(r, password)

        return await self.invoke(
            raw.functions.messages.GetBotCallbackAnswer(
                peer=await self.resolve_peer(chat_id),
                msg_id=message_id,
                data=data,
                password=password
            ),
            retries=0,
            timeout=timeout
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging

import Badmunda
from Badmunda import raw
from Badmunda.raw.core import TLObject
from Badmunda.session import Session
from Badmunda.methods.messages.business_session import get_session

log = logging.getLogger(__name__)


class Invoke:
    async def invoke(
        self: "Badmunda.Client",
        query: TLObject,
        retries: int = Session.MAX_RETRIES,
        timeout: float = Session.WAIT_TIMEOUT,
        sleep_threshold: float = None,
        business_connection_id: str = None
    ):
        """Invoke raw Telegram functions.

        This method makes it possible to manually call every single Telegram API method in a low-level manner.
        Available functions are listed in the :obj:`functions <Badmunda.api.functions>` package and may accept compound
        data types from :obj:`types <Badmunda.api.types>` as well as bare types such as ``int``, ``str``, etc...

        .. note::

            This is a utility method intended to be used **only** when working with raw
            :obj:`functions <Badmunda.api.functions>` (i.e: a Telegram API method you wish to use which is not
            available yet in the Client class as an easy-to-use method).

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            query (``RawFunction``):
                The API Schema function filled with proper arguments.

            retries (``int``):
                Number of retries.

            timeout (``float``):
                Timeout in seconds.

            sleep_threshold (``float``):
                Sleep threshold in seconds.

            business_connection_id (``str``, *optional*):
                Unique identifier of the business connection.

        Returns:
            ``RawType``: The raw type response generated by the query.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        if not self.is_connected:
            raise ConnectionError("Client has not been started yet")

        session = self.session

        if business_connection_id:
            query = raw.functions.InvokeWithBusinessConnection(
                connection_id=business_connection_id,
                query=query
            )

            session = await get_session(self, business_connection_id)

        if self.no_updates:
            query = raw.functions.InvokeWithoutUpdates(query=query)

        if self.takeout_id:
            query = raw.functions.InvokeWithTakeout(takeout_id=self.takeout_id, query=query)

        r = await session.invoke(
            query, retries, timeout,
            (sleep_threshold
             if sleep_threshold is not None
             else self.sleep_threshold)
        )

        await self.fetch_peers(getattr(r, "users", []))
        await self.fetch_peers(getattr(r, "chats", []))

        return r



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import List

import Badmunda
from Badmunda import raw
from Badmunda import types

log = logging.getLogger(__name__)


class GetStickers:
    async def get_stickers(
        self: "Badmunda.Client",
        short_name: str
    ) -> List["types.Sticker"]:
        """Get all stickers from set by short name.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            short_name (``str``):
                Short name of the sticker set, serves as the unique identifier for the sticker set.

        Returns:
            List of :obj:`~Badmunda.types.Sticker`: A list of stickers is returned.

        Example:
            .. code-block:: python

                # Get all stickers by short name
                await app.get_stickers("short_name")

        Raises:
            ValueError: In case of invalid arguments.
        """
        sticker_set = await self.invoke(
            raw.functions.messages.GetStickerSet(
                stickerset=raw.types.InputStickerSetShortName(short_name=short_name),
                hash=0
            )
        )

        return types.List(
            [
                await types.Sticker._parse(self, doc, {type(a): a for a in doc.attributes})
                for doc in sticker_set.documents
            ]
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional, Union

import Badmunda
from Badmunda import enums, raw, types, utils


class CreateFolder:
    async def create_folder(
        self: "Badmunda.Client",
        name: str,
        parse_mode: Optional["enums.ParseMode"] = None,
        entities: Optional[List["types.MessageEntity"]] = None,
        animate_custom_emoji: Optional[bool] = None,
        icon: Optional[str] = None,
        color: Optional["enums.FolderColor"] = None,
        pinned_chats: Optional[List[Union[int, str]]] = None,
        included_chats: Optional[List[Union[int, str]]] = None,
        excluded_chats: Optional[List[Union[int, str]]] = None,
        exclude_muted: Optional[bool] = None,
        exclude_read: Optional[bool] = None,
        exclude_archived: Optional[bool] = None,
        include_contacts: Optional[bool] = None,
        include_non_contacts: Optional[bool] = None,
        include_bots: Optional[bool] = None,
        include_groups: Optional[bool] = None,
        include_channels: Optional[bool] = None
    ) -> int:
        """Create new chat folder.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            name (``str``):
                The text of the chat folder name, 1-12 characters without line feeds.

            parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            entities (List of :obj:`~Badmunda.types.MessageEntity`, *optional*):
                Special entities like bold, italic, etc. that appear in the folder name.

            animate_custom_emoji (``bool``, *optional*):
                True, if custom emoji in the name must be animated.

            icon (``str``, *optional*):
                The chosen icon for the chat folder.
                Pass None to leave the folder icon as default.

            color (:obj:`~Badmunda.enums.FolderColor`, *optional*)
                Pass :obj:`~Badmunda.enums.FolderColor` to set folder color.
                Can't be changed if folder tags are disabled or the current user doesn't have Telegram Premium subscription.

            is_shareable (``bool``, *optional*):
                True, if at least one link has been created for the folder.

            pinned_chats (List of :obj:`~Badmunda.types.Chat`, *optional*):
                The pinned chats in the folder.
                You can pass an ID (int), username (str) or phone number (str).
                There can be up to getOption("chat_folder_chosen_chat_count_max") pinned and always included non-secret chats and the same number of secret chats, but the limit can be increased with Telegram Premium.

            included_chats (List of :obj:`~Badmunda.types.Chat`, *optional*):
                The always included chats in the folder.
                You can pass an ID (int), username (str) or phone number (str).
                There can be up to getOption("chat_folder_chosen_chat_count_max") pinned and always included non-secret chats and the same number of secret chats, but the limit can be increased with Telegram Premium.

            excluded_chats (List of :obj:`~Badmunda.types.Chat`, *optional*):
                The always excluded chats in the folder.
                You can pass an ID (int), username (str) or phone number (str).
                There can be up to getOption("chat_folder_chosen_chat_count_max") always excluded non-secret chats and the same number of secret chats, but the limit can be increased with Telegram Premium.

            exclude_muted (``bool``, *optional*):
                True, if muted chats need to be excluded.

            exclude_read (``bool``, *optional*):
                True, if read chats need to be excluded.

            exclude_archived (``bool``, *optional*):
                True, if archived chats need to be excluded.

            include_contacts (``bool``, *optional*):
                True, if contacts need to be included.

            include_non_contacts (``bool``, *optional*):
                True, if non-contact users need to be included.

            include_bots (``bool``, *optional*):
                True, if bots need to be included.

            include_groups (``bool``, *optional*):
                True, if basic groups and supergroups need to be included.

            include_channels (``bool``, *optional*):
                True, if channels need to be included.

        Returns:
            ``int``: On success, folder id is returned.

        Example:
            .. code-block:: python

                # Create folder
                await app.create_folder(name="New folder", included_chats=["me"])
        """
        dialog_filters = await self.invoke(raw.functions.messages.GetDialogFilters())

        raw_folders_ids = [
            folder.id for folder in dialog_filters.filters
            if isinstance(folder, (raw.types.DialogFilter, raw.types.DialogFilterChatlist))
        ]

        # find first free folder id
        for i in range(2, 256):
            if i not in set(raw_folders_ids):
                folder_id = i
                break

        name, title_entities = (await utils.parse_text_entities(self, name, parse_mode, entities)).values()
        title_entities = title_entities or []

        pinned_chats = pinned_chats or []
        included_chats = included_chats or []
        excluded_chats = excluded_chats or []

        await self.invoke(
            raw.functions.messages.UpdateDialogFilter(
                id=folder_id,
                filter=raw.types.DialogFilter(
                    id=folder_id,
                    title=raw.types.TextWithEntities(
                        text=name,
                        entities=title_entities,
                    ),
                    pinned_peers=[
                        await self.resolve_peer(user_id)
                        for user_id in pinned_chats
                    ],
                    include_peers=[
                        await self.resolve_peer(user_id)
                        for user_id in included_chats
                    ],
                    exclude_peers=[
                        await self.resolve_peer(user_id)
                        for user_id in excluded_chats
                    ],
                    contacts=include_contacts,
                    non_contacts=include_non_contacts,
                    groups=include_groups,
                    broadcasts=include_channels,
                    bots=include_bots,
                    exclude_muted=exclude_muted,
                    exclude_read=exclude_read,
                    exclude_archived=exclude_archived,
                    title_noanimate=not animate_custom_emoji,
                    emoticon=icon,
                    color=color.value if color else None
                )
            )
        )

        return folder_id



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

import Badmunda
from Badmunda import raw, types


class SetGlobalPrivacySettings:
    async def set_global_privacy_settings(
        self: "Badmunda.Client",
        archive_and_mute_new_chats: Optional[bool] = None,
        keep_unmuted_chats_archived: Optional[bool] = None,
        keep_chats_from_folders_archived: Optional[bool] = None,
        show_read_date: Optional[bool] = None,
        allow_new_chats_from_unknown_users: Optional[bool] = None,
        incoming_paid_message_star_count: Optional[int] = None,
        show_gift_button: Optional[bool] = None,
        accepted_gift_types: Optional[types.AcceptedGiftTypes] = None,
    ) -> "types.GlobalPrivacySettings":
        """Set account global privacy settings.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            archive_and_mute_new_chats (``bool``, *optional*):
                Whether to archive and mute new chats from non-contacts,

            keep_unmuted_chats_archived (``bool``, *optional*):
                Whether unmuted chats will be kept in
                the Archive chat list when they get a new message.

            keep_chats_from_folders_archived (``bool``, *optional*):
                Whether unmuted chats, that are always included or pinned
                in a folder, will be kept in the Archive chat list when they get
                a new message. Ignored if keep_unmuted_chats_archived is set.

            show_read_date (``bool``, *optional*):
                True, if message read date is shown to other users in private chats.
                If false and the current user isn't a Telegram Premium user,
                then they will not be able to see other's message read date.

            allow_new_chats_from_unknown_users (``bool``, *optional*):
                True, if non-contacts users are able to write first to the current user.
                Telegram Premium subscribers are able to write first regardless of this setting.

            incoming_paid_message_star_count (``int``, *optional*):
                Number of Telegram Stars that must be paid for every incoming private message
                by non-contacts.
                Must be between 0-``paid_message_star_count_max``.
                If positive, then allow_new_chats_from_unknown_users must be true.
                The current user will receive ``paid_message_earnings_per_mille`` Telegram Stars
                for each 1000 Telegram Stars paid for message sending.

            show_gift_button (``bool``, *optional*):
                True, if a button for sending a gift to the user or by the user must be always shown in the input field.

            accepted_gift_types (:obj:`~Badmunda.types.AcceptedGiftTypes`, *optional*):
                Information about gifts that can be received by the user.

        Returns:
            :obj:`~Badmunda.types.GlobalPrivacySettings`: On success, the new global privacy settings is returned.

        Example:
            .. code-block:: python

                # Archive new chats
                await app.set_global_privacy_settings(
                    archive_and_mute_new_chats=True
                )

                # Set price for incoming messages
                await app.set_global_privacy_settings(
                    incoming_paid_message_star_count=10
                )
        """
        settings = await self.invoke(raw.functions.account.GetGlobalPrivacySettings())

        if archive_and_mute_new_chats is not None:
            settings.archive_and_mute_new_noncontact_peers = archive_and_mute_new_chats

        if keep_unmuted_chats_archived is not None:
            settings.keep_archived_unmuted = keep_unmuted_chats_archived

        if keep_chats_from_folders_archived is not None:
            settings.keep_archived_folders = keep_chats_from_folders_archived

        if show_read_date is not None:
            settings.hide_read_marks = show_read_date

        if allow_new_chats_from_unknown_users is not None:
            settings.new_noncontact_peers_require_premium = allow_new_chats_from_unknown_users

        if incoming_paid_message_star_count is not None:
            settings.noncontact_peers_paid_stars = incoming_paid_message_star_count

        if show_gift_button is not None:
            settings.show_gift_button = show_gift_button

        if accepted_gift_types is not None:
            settings.disallowed_gifts = accepted_gift_types.write()

        r = await self.invoke(
            raw.functions.account.SetGlobalPrivacySettings(
                settings=settings
            )
        )

        return types.GlobalPrivacySettings._parse(r)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from Badmunda import raw, utils
from ..object import Object


class VideoChatScheduled(Object):
    """A service message about a voice chat scheduled in the chat.

    Parameters:
        start_date (:py:obj:`~datetime.datetime`):
            Point in time when the voice chat is supposed to be started by a chat administrator.
    """

    def __init__(
        self, *,
        start_date: datetime
    ):
        super().__init__()

        self.start_date = start_date

    @staticmethod
    def _parse(action: "raw.types.MessageActionGroupCallScheduled") -> "VideoChatScheduled":
        return VideoChatScheduled(start_date=utils.timestamp_to_datetime(action.schedule_date))



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from Badmunda import raw

from ..object import Object


class LoginUrl(Object):
    """Represents a parameter of the inline keyboard button used to automatically authorize a user.

    Serves as a great replacement for the Telegram Login Widget when the user is coming from Telegram.
    All the user needs to do is tap/click a button and confirm that they want to log in.

    Parameters:
        url (``str``):
            An HTTP URL to be opened with user authorization data added to the query string when the button is pressed.
            If the user refuses to provide authorization data, the original URL without information about the user will
            be opened. The data added is the same as described in
            `Receiving authorization data <https://core.telegram.org/widgets/login#receiving-authorization-data>`.

            **NOTE**: You **must** always check the hash of the received data to verify the authentication and the
            integrity of the data as described in
            `Checking authorization <https://core.telegram.org/widgets/login#checking-authorization>`_.

        forward_text (``str``, *optional*):
            New text of the button in forwarded messages.

        bot_username (``str``, *optional*):
            Username of a bot, which will be used for user authorization.
            See `Setting up a bot <https://core.telegram.org/widgets/login#setting-up-a-bot>`_ for more details.
            If not specified, the current bot's username will be assumed. The url's domain must be the same as the
            domain linked with the bot.
            See `Linking your domain to the bot <https://core.telegram.org/widgets/login#linking-your-domain-to-the-bot>`_
            for more details.

        request_write_access (``str``, *optional*):
            Pass True to request the permission for your bot to send messages to the user.

        button_id (``int``):
            Button identifier.
    """

    def __init__(
        self, *,
        url: str,
        forward_text: str = None,
        bot_username: str = None,
        request_write_access: str = None,
        button_id: int = None
    ):
        super().__init__()

        self.url = url
        self.forward_text = forward_text
        self.bot_username = bot_username
        self.request_write_access = request_write_access
        self.button_id = button_id

    @staticmethod
    def read(b: "raw.types.KeyboardButtonUrlAuth") -> "LoginUrl":
        return LoginUrl(
            url=b.url,
            forward_text=b.fwd_text,
            button_id=b.button_id
        )

    def write(self, text: str, bot: "raw.types.InputUser"):
        return raw.types.InputKeyboardButtonUrlAuth(
            text=text,
            url=self.url,
            bot=bot,
            fwd_text=self.forward_text,
            request_write_access=self.request_write_access
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import io
import os
import re
from typing import Union

import Badmunda
from Badmunda import raw
from Badmunda import types
from Badmunda import utils
from Badmunda.file_id import FileType


class EditMessageMedia:
    async def edit_message_media(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        message_id: int,
        media: "types.InputMedia",
        show_caption_above_media: bool = None,
        schedule_date: datetime = None,
        reply_markup: "types.InlineKeyboardMarkup" = None,
        file_name: str = None
    ) -> "types.Message":
        """Edit animation, audio, document, photo or video messages, or to add media to text messages.

        If a message is part of a message album, then it can be edited only to an audio for audio albums, only to a document for document albums and to a photo or a video otherwise.
        Otherwise, the message type can be changed arbitrarily.

        .. note::

           Business messages that were not sent by the bot and do not contain an inline keyboard can only be edited within 48 hours from the time they were sent.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            message_id (``int``):
                Message identifier in the chat specified in chat_id.

            media (:obj:`~Badmunda.types.InputMedia`):
                One of the InputMedia objects describing an animation, audio, document, photo or video.

            show_caption_above_media (``bool``, *optional*):
                Pass True, if the caption must be shown above the message media.

            schedule_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the message will be automatically sent.

            reply_markup (:obj:`~Badmunda.types.InlineKeyboardMarkup`, *optional*):
                An InlineKeyboardMarkup object.

            file_name (``str``, *optional*):
                File name of the media to be sent. Not applicable to photos.
                Defaults to file's path basename.

        Returns:
            :obj:`~Badmunda.types.Message`: On success, the edited message is returned.

        Example:
            .. code-block:: python

                from Badmunda.types import InputMediaPhoto, InputMediaVideo, InputMediaAudio

                # Replace the current media with a local photo
                await app.edit_message_media(chat_id, message_id,
                    InputMediaPhoto("new_photo.jpg"))

                # Replace the current media with a local video
                await app.edit_message_media(chat_id, message_id,
                    InputMediaVideo("new_video.mp4"))

                # Replace the current media with a local audio
                await app.edit_message_media(chat_id, message_id,
                    InputMediaAudio("new_audio.mp3"))
        """
        caption = media.caption
        parse_mode = media.parse_mode
        caption_entities = media.caption_entities

        message, entities = None, None

        if caption is not None:
            message, entities = (await utils.parse_text_entities(self, caption, parse_mode, caption_entities)).values()

        if isinstance(media, types.InputMediaPhoto):
            if isinstance(media.media, io.BytesIO) or os.path.isfile(media.media):
                uploaded_media = await self.invoke(
                    raw.functions.messages.UploadMedia(
                        peer=await self.resolve_peer(chat_id),
                        media=raw.types.InputMediaUploadedPhoto(
                            file=await self.save_file(media.media),
                            spoiler=media.has_spoiler
                        )
                    )
                )

                media = raw.types.InputMediaPhoto(
                    id=raw.types.InputPhoto(
                        id=uploaded_media.photo.id,
                        access_hash=uploaded_media.photo.access_hash,
                        file_reference=uploaded_media.photo.file_reference
                    ),
                    spoiler=media.has_spoiler
                )
            elif re.match("^https?://", media.media):
                media = raw.types.InputMediaPhotoExternal(
                    url=media.media,
                    spoiler=media.has_spoiler
                )
            else:
                media = utils.get_input_media_from_file_id(media.media, FileType.PHOTO, has_spoiler=media.has_spoiler)
        elif isinstance(media, types.InputMediaVideo):
            if isinstance(media.media, io.BytesIO) or os.path.isfile(media.media):
                uploaded_media = await self.invoke(
                    raw.functions.messages.UploadMedia(
                        peer=await self.resolve_peer(chat_id),
                        media=raw.types.InputMediaUploadedDocument(
                            mime_type=self.guess_mime_type(media.media) or "video/mp4",
                            thumb=await self.save_file(media.thumb),
                            spoiler=media.has_spoiler,
                            file=await self.save_file(media.media),
                            attributes=[
                                raw.types.DocumentAttributeVideo(
                                    supports_streaming=media.supports_streaming or None,
                                    duration=media.duration,
                                    w=media.width,
                                    h=media.height
                                ),
                                raw.types.DocumentAttributeFilename(
                                    file_name=file_name or os.path.basename(media.media)
                                )
                            ]
                        )
                    )
                )

                media = raw.types.InputMediaDocument(
                    id=raw.types.InputDocument(
                        id=uploaded_media.document.id,
                        access_hash=uploaded_media.document.access_hash,
                        file_reference=uploaded_media.document.file_reference
                    ),
                    spoiler=media.has_spoiler
                )
            elif re.match("^https?://", media.media):
                media = raw.types.InputMediaDocumentExternal(
                    url=media.media,
                    spoiler=media.has_spoiler
                )
            else:
                media = utils.get_input_media_from_file_id(media.media, FileType.VIDEO, has_spoiler=media.has_spoiler)
        elif isinstance(media, types.InputMediaAudio):
            if isinstance(media.media, io.BytesIO) or os.path.isfile(media.media):
                media = await self.invoke(
                    raw.functions.messages.UploadMedia(
                        peer=await self.resolve_peer(chat_id),
                        media=raw.types.InputMediaUploadedDocument(
                            mime_type=self.guess_mime_type(media.media) or "audio/mpeg",
                            thumb=await self.save_file(media.thumb),
                            file=await self.save_file(media.media),
                            attributes=[
                                raw.types.DocumentAttributeAudio(
                                    duration=media.duration,
                                    performer=media.performer,
                                    title=media.title
                                ),
                                raw.types.DocumentAttributeFilename(
                                    file_name=file_name or os.path.basename(media.media)
                                )
                            ]
                        )
                    )
                )

                media = raw.types.InputMediaDocument(
                    id=raw.types.InputDocument(
                        id=media.document.id,
                        access_hash=media.document.access_hash,
                        file_reference=media.document.file_reference
                    )
                )
            elif re.match("^https?://", media.media):
                media = raw.types.InputMediaDocumentExternal(
                    url=media.media
                )
            else:
                media = utils.get_input_media_from_file_id(media.media, FileType.AUDIO)
        elif isinstance(media, types.InputMediaAnimation):
            if isinstance(media.media, io.BytesIO) or os.path.isfile(media.media):
                uploaded_media = await self.invoke(
                    raw.functions.messages.UploadMedia(
                        peer=await self.resolve_peer(chat_id),
                        media=raw.types.InputMediaUploadedDocument(
                            mime_type=self.guess_mime_type(media.media) or "video/mp4",
                            thumb=await self.save_file(media.thumb),
                            spoiler=media.has_spoiler,
                            file=await self.save_file(media.media),
                            attributes=[
                                raw.types.DocumentAttributeVideo(
                                    supports_streaming=True,
                                    duration=media.duration,
                                    w=media.width,
                                    h=media.height
                                ),
                                raw.types.DocumentAttributeFilename(
                                    file_name=file_name or os.path.basename(media.media)
                                ),
                                raw.types.DocumentAttributeAnimated()
                            ]
                        )
                    )
                )

                media = raw.types.InputMediaDocument(
                    id=raw.types.InputDocument(
                        id=uploaded_media.document.id,
                        access_hash=uploaded_media.document.access_hash,
                        file_reference=uploaded_media.document.file_reference
                    ),
                    spoiler=media.has_spoiler
                )
            elif re.match("^https?://", media.media):
                media = raw.types.InputMediaDocumentExternal(
                    url=media.media,
                    spoiler=media.has_spoiler
                )
            else:
                media = utils.get_input_media_from_file_id(media.media, FileType.ANIMATION, has_spoiler=media.has_spoiler)
        elif isinstance(media, types.InputMediaDocument):
            if isinstance(media.media, io.BytesIO) or os.path.isfile(media.media):
                media = await self.invoke(
                    raw.functions.messages.UploadMedia(
                        peer=await self.resolve_peer(chat_id),
                        media=raw.types.InputMediaUploadedDocument(
                            mime_type=self.guess_mime_type(media.media) or "application/zip",
                            thumb=await self.save_file(media.thumb),
                            file=await self.save_file(media.media),
                            attributes=[
                                raw.types.DocumentAttributeFilename(
                                    file_name=file_name or os.path.basename(media.media)
                                )
                            ]
                        )
                    )
                )

                media = raw.types.InputMediaDocument(
                    id=raw.types.InputDocument(
                        id=media.document.id,
                        access_hash=media.document.access_hash,
                        file_reference=media.document.file_reference
                    )
                )
            elif re.match("^https?://", media.media):
                media = raw.types.InputMediaDocumentExternal(
                    url=media.media
                )
            else:
                media = utils.get_input_media_from_file_id(media.media, FileType.DOCUMENT)

        r = await self.invoke(
            raw.functions.messages.EditMessage(
                peer=await self.resolve_peer(chat_id),
                id=message_id,
                invert_media=show_caption_above_media,
                media=media,
                schedule_date=utils.datetime_to_timestamp(schedule_date),
                reply_markup=await reply_markup.write(self) if reply_markup else None,
                message=message,
                entities=entities
            )
        )

        for i in r.updates:
            if isinstance(i, (raw.types.UpdateEditMessage, raw.types.UpdateEditChannelMessage)):
                return await types.Message._parse(
                    self, i.message,
                    {i.id: i for i in r.users},
                    {i.id: i for i in r.chats}
                )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional
from Badmunda import raw

from ..object import Object


class ShippingAddress(Object):
    """Contains information about a shipping address.

    Parameters:
        country_code (``str``):
            Two-letter `ISO 3166-1 alpha-2 <https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2>`_ country code.

        state (``str``):
            State, if applicable.

        city (``str``):
            City.

        street_line1 (``str``):
            First line for the address.

        street_line2 (``str``):
            Second line for the address.

        post_code (``str``):
            Address post code.

    """

    def __init__(
        self,
        *,
        country_code: str,
        state: str,
        city: str,
        street_line1: str,
        street_line2: str,
        post_code: str
    ):
        super().__init__()

        self.country_code = country_code
        self.state = state
        self.city = city
        self.street_line1 = street_line1
        self.street_line2 = street_line2
        self.post_code = post_code

    @staticmethod
    def _parse(
        shipping_address: "raw.types.PostAddress",
    ) -> Optional["ShippingAddress"]:
        if not shipping_address:
            return None

        return ShippingAddress(
            country_code=shipping_address.country_iso2,
            state=shipping_address.state,
            city=shipping_address.city,
            street_line1=shipping_address.street_line1,
            street_line2=shipping_address.street_line2,
            post_code=shipping_address.post_code
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import Badmunda
from Badmunda import raw


class GetBusinessAccountStarBalance:
    async def get_business_account_star_balance(
        self: "Badmunda.Client",
        business_connection_id: str,
    ) -> int:
        """Return the amount of Telegram Stars owned by a managed business account.

        .. note::

            Requires the `can_view_gifts_and_stars` business bot right.

        .. include:: /_includes/usable-by/bots.rst

        Parameters:
            business_connection_id (``str``):
                Unique identifier of business connection on behalf of which to send the request.

        Returns:
            ``int``: On success, the current stars balance is returned.

        Example:
            .. code-block:: python

                # Get stars balance
                await app.get_business_account_star_balance("connection_id")
        """
        connection_info = await self.get_business_connection(business_connection_id)

        r = await self.invoke(
            raw.functions.payments.GetStarsStatus(
                peer=await self.resolve_peer(connection_info.user.id),
            ),
            business_connection_id=business_connection_id
        )

        return r.balance.amount



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging
from datetime import datetime
from typing import AsyncGenerator, BinaryIO, List, Optional, Union

import Badmunda
from Badmunda import enums, raw, types, utils

from ..object import Object

log = logging.getLogger(__name__)


class Chat(Object):
    """A chat.

    Parameters:
        id (``int``, *optional*):
            Unique identifier for this chat.

        type (:obj:`~Badmunda.enums.ChatType`, *optional*):
            Type of chat.

        is_forum (``bool``, *optional*):
            True, if the supergroup chat is a forum.

        is_direct_messages_group (``bool``, *optional*):
            True, if the supergroup is a direct message group for a channel chat.

        is_min (``bool``, *optional*):
            True, if this chat have reduced set of fields.

        is_members_hidden (``bool``, *optional*):
            True, if the chat members are hidden.

        is_restricted (``bool``, *optional*):
            True, if this chat has been restricted. Supergroups, channels and bots only.
            See *restriction_reason* for details.

        is_creator (``bool``, *optional*):
            True, if this chat owner is the current user. Supergroups, channels and groups only.

        is_admin (``bool``, *optional*):
            True, if the current user is admin. Supergroups, channels and groups only.

        is_deactivated (``bool``, *optional*):
            True, if this chat has been flagged for deactivated.

        is_support (``bool``, *optional*):
            True, if this chat is part of the Telegram support team. Users and bots only.

        is_stories_hidden (``bool``, *optional*):
            True, if this chat has hidden stories.

        is_stories_unavailable (``bool``, *optional*):
            True, if this chat stories is unavailable.

        is_business_bot (``bool``, *optional*):
            True, if this bot can connect to business account.

        is_preview (``bool``, *optional*):
            True, if this chat is a preview.

        is_banned (``bool``, *optional*):
            True, if you are banned in this chat.

        is_call_active (``bool``, *optional*):
            True, if a group call is currently active.

        is_call_not_empty (``bool``, *optional*):
            True, if there's anyone in the group call.

        is_public (``bool``, *optional*):
            True, if this chat is public.

        is_paid_reactions_available (``bool``, *optional*):
            True, if paid reactions enabled in this chat.

        verification_status (:obj:`~Badmunda.types.VerificationStatus`, *optional*):
            Contains information about verification status of a chat.

        can_send_gift (``bool``, *optional*):
            True, if the user can send a gift to the supergroup or channel using :meth:`~Badmunda.Client.send_gift` or :meth:`~Badmunda.Client.transfer_gift`.

        title (``str``, *optional*):
            Title, for supergroups, channels and basic group chats.

        username (``str``, *optional*):
            Username, for private chats, bots, supergroups and channels if available.

        usernames (List of :obj:`~Badmunda.types.Username`, *optional*):
            The list of chat's collectible (and basic) usernames if available.

        first_name (``str``, *optional*):
            First name of the other party in a private chat, for private chats and bots.

        last_name (``str``, *optional*):
            Last name of the other party in a private chat, for private chats.

        photo (:obj:`~Badmunda.types.ChatPhoto`, *optional*):
            Chat photo. Suitable for downloads only.

        stories (List of :obj:`~Badmunda.types.Story`, *optional*):
            The list of chat's stories if available.

        chat_background (:obj:`~Badmunda.types.ChatBackground`, *optional*):
            Chat wallpaper.

        bio (``str``, *optional*):
            Bio of the other party in a private chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        description (``str``, *optional*):
            Description, for groups, supergroups and channel chats.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        show_message_sender_name (``bool``, *optional*):
            True, if the chat has a username.

        sign_messages (``bool``, *optional*):
            True, if messages sent to the channel contains name of the sender. This field is only applicable to channels.

        dc_id (``int``, *optional*):
            The chat assigned DC (data center). Available only in case the chat has a photo.
            Note that this information is approximate; it is based on where Telegram stores the current chat photo.
            It is accurate only in case the owner has set the chat photo, otherwise the dc_id will be the one assigned
            to the administrator who set the current chat photo.

        folder_id (``int``, *optional*):
            The folder identifier where the chat is located.

        has_protected_content (``bool``, *optional*):
            True, if messages from the chat can't be forwarded to other chats.

        has_visible_history (``bool``, *optional*):
            True, if new chat members will have access to old messages; available only to chat administrators.

        has_aggressive_anti_spam_enabled (``bool``, *optional*):
            True, if aggressive anti-spam checks are enabled in the supergroup. The field is only available to chat administrators.

        has_automatic_translation (``bool``, *optional*):
            True, if automatic translation of messages is enabled in the channel.

        has_forum_tabs (``bool``, *optional*):
            True, if the supergroup is a forum, which topics are shown in the same way as in channel direct messages groups.

        has_direct_messages_group (``bool``, *optional*):
            True, if the channel has direct messages group.

        direct_messages_chat_id (``int``, *optional*):
            Chat identifier of a direct messages group for the channel, or a channel, for which the supergroup is the designated direct messages group.

        invite_link (``str``, *optional*):
            Chat invite link, for groups, supergroups and channels.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        pinned_message (:obj:`~Badmunda.types.Message`, *optional*):
            Pinned message, for groups, supergroups channels and own chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        sticker_set_name (``str``, *optional*):
            For supergroups, name of group sticker set.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        custom_emoji_sticker_set_name (``str``, *optional*):
            For supergroups, the name of the group's custom emoji sticker set. Custom emoji from this set can be used by all users and bots in the group.

        can_set_sticker_set (``bool``, *optional*):
            True, if the group sticker set can be changed by you.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        can_send_paid_media (``bool``, *optional*):
            True, if paid media messages can be sent or forwarded to the channel chat. The field is available only for channel chats.

        members (List of :obj:`~Badmunda.types.User`, *optional*):
            A few of the participants that are in the group.

        members_count (``int``, *optional*):
            Chat members count, for groups, supergroups and channels only.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        restrictions (List of :obj:`~Badmunda.types.Restriction`, *optional*):
            The list of reasons why this chat might be unavailable to some users.
            This field is available only in case *is_restricted* is True.

        permissions (:obj:`~Badmunda.types.ChatPermissions` *optional*):
            Default chat member permissions, for groups and supergroups.

        personal_channel (:obj:`~Badmunda.types.Chat`, *optional*):
            The personal channel linked to this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        personal_channel_message (:obj:`~Badmunda.types.Message`, *optional*):
            The last message in the personal channel of this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        linked_chat (:obj:`~Badmunda.types.Chat`, *optional*):
            The linked discussion group (in case of channels) or the linked channel (in case of supergroups).
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        send_as_chat (:obj:`~Badmunda.types.Chat`, *optional*):
            The default "send_as" chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        available_reactions (:obj:`~Badmunda.types.ChatReactions`, *optional*):
            Available reactions in the chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        level (``int``, *optional*):
            Channel boosts level.

        reply_color (:obj:`~Badmunda.types.ChatColor`, *optional*):
            Chat reply color.

        profile_color (:obj:`~Badmunda.types.ChatColor`, *optional*):
            Chat profile color.

        business_away_message (:obj:`~Badmunda.types.BusinessMessage`, *optional*):
            For private chats with business accounts, the away message of the business.

        business_greeting_message (:obj:`~Badmunda.types.BusinessMessage`, *optional*):
            For private chats with business accounts, the greeting message of the business.

        business_work_hours (:obj:`~Badmunda.types.BusinessWorkingHours`, *optional*):
            For private chats with business accounts, the working hours of the business.

        business_location (:obj:`~Badmunda.types.Location`, *optional*):
            For private chats with business accounts, the location of the business.

        business_intro (:obj:`~Badmunda.types.BusinessIntro`, *optional*):
            For private chats with business accounts, the intro of the business.

        birthday (:obj:`~Badmunda.types.Birthday`, *optional*):
            Information about user birthday.

        message_auto_delete_time (``int``, *optional*):
            The time after which all messages sent to the chat will be automatically deleted; in seconds.

        unrestrict_boost_count (``int``, *optional*):
            For supergroups, the minimum number of boosts that a non-administrator user needs to add in order to ignore slow mode and chat permissions.

        slow_mode_delay (``int``, *optional*):
            For supergroups, the minimum allowed delay between consecutive messages sent by each unpriviledged user; in seconds.

        slowmode_next_send_date (:py:obj:`~datetime.datetime`, *optional*):
            Indicates when the user will be allowed to send another message in the chat. For supergroups only.

        join_by_request (``bool``, *optional*):
            True, if all users directly joining the supergroup need to be approved by supergroup administrators.

        join_requests_count (``int``, *optional*):
            Number of users who requested to join the chat.

        banned_until_date (:py:obj:`~datetime.datetime`, *optional*):
            Date when the user will be unbanned.

        subscription_until_date (:py:obj:`~datetime.datetime`, *optional*):
            Date when the the subscription will end.

        reactions_limit (``int``, *optional*):
            This flag may be used to impose a custom limit of unique reactions (i.e. a customizable version of appConfig.reactions_uniq_max).

        gift_count (``int``, *optional*):
            Number of saved to profile gifts for channels without `can_post_messages` administrator right, otherwise, the total number of received gifts.

        bot_verification (:obj:`~Badmunda.types.BotVerification`, *optional*):
            Information about bot verification.

        settings (:obj:`~Badmunda.types.ChatSettings`, *optional*):
            Chat settings.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        admins_count (``int``, *optional*):
            Number of admins in channel.
            Returned only in :meth:`~Badmunda.Client.get_chat`.

        kicked_count (``int``, *optional*):
            Number of kicked from the channel.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        banned_count (``int``, *optional*):
            Number of banned from the channel.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        available_min_id (``int``, *optional*):
            Identifier of a maximum unavailable message due to hidden history.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        boosts_applied (``int``, *optional*):
            The number of boosts the current user has applied to the current supergroup.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        bot_broadcast_admin_rights (:obj:`~Badmunda.types.ChatPrivileges`, *optional*):
            A suggested set of administrator rights for the bot, to be shown when adding the bot as admin to a channel.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        bot_group_admin_rights (:obj:`~Badmunda.types.ChatPrivileges`, *optional*):
            A suggested set of administrator rights for the bot, to be shown when adding the bot as admin to a group.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        bot_can_manage_emoji_status (``bool``, *optional*):
            True, if the bot can change your emoji status.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_delete_channel (``bool``, *optional*):
            True, if the current user can delete this channel.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_pin_message (``bool``, *optional*):
            True, if the current user can pin messages in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_schedule_messages (``bool``, *optional*):
            True, if the current user can schedule messages in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_set_location (``bool``, *optional*):
            True, if the current user can set location in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_set_username (``bool``, *optional*):
            True, if the current user can set username in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_view_participants (``bool``, *optional*):
            True, if the current user can view participants in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_view_revenue (``bool``, *optional*):
            True, if the current user can view revenue in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_view_stars_revenue (``bool``, *optional*):
            True, if the current user can view stars revenue in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        can_view_stats (``bool``, *optional*):
            True, if the current user can view stats in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        common_chats (``int``, *optional*):
            Number of common chats with this user.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_ads_enabled (``bool``, *optional*):
            True, if ads were re-enabled for the current account (only accessible to the currently logged-in user).
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_blocked (``bool``, *optional*):
            True, if you have blocked this user.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_blocked_my_stories_from (``bool``, *optional*):
            True, if we've blocked this user, preventing them from seeing our stories.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_contact_require_premium (``bool``, *optional*):
            True, if we cannot write to this user:
            subscribe to Telegram Premium to get permission to write to this user.
            To set this flag for ourselves invoke account.setGlobalPrivacySettings,
            setting the settings.new_noncontact_peers_require_premium flag.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_phone_calls_available (``bool``, *optional*):
            True, if this user can make VoIP calls.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_phone_calls_private (``bool``, *optional*):
            True, if this user's privacy settings allow you to call them.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_pinned_stories_available (``bool``, *optional*):
            True, if this user has some pinned stories.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_read_dates_available (``bool``, *optional*):
            True, if we cannot fetch the exact read date of messages we send to this user.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_translations_disabled (``bool``, *optional*):
            True, if the real-time chat translation popup should be hidden.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_video_calls_available (``bool``, *optional*):
            True, if this user can receive video calls.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_wallpaper_overridden (``bool``, *optional*):
            True, if this user has a custom wallpaper.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        migrated_from_chat_id (``int``, *optional*):
            The unique chat identifier from which this group was migrated.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        migrated_from_max_message_id (``int``, *optional*):
            The message identifier in the original chat at which this group was migrated.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        online_count (``int``, *optional*):
            Number of online members in the chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        private_forward_name (``str``, *optional*):
            Anonymized text to be shown instead of the user's name on forwarded messages.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        read_inbox_max_id (``int``, *optional*):
            Position up to which all incoming messages are read.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        read_outbox_max_id (``int``, *optional*):
            Position up to which all outgoing messages are read.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        is_ads_restricted (``bool``, *optional*):
            True, if ads on this channel were restricted.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        stats_dc_id (``int``, *optional*):
            The DC ID where the stats are stored.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        theme_emoji (``str``, *optional*):
            Emoji representing a specific chat theme.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        unread_count (``int``, *optional*):
            Number of unread messages in the chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        view_forum_as_messages (``bool``, *optional*):
            Users may also choose to display messages from all topics of a forum as if they were sent to a normal group,
            using a "View as messages" setting in the local client.
            This setting only affects the current account, and is synced to other logged in sessions using the channels.toggleViewForumAsMessages method.
            Invoking this method will update the value of this flag.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        paid_message_star_count (``int``, *optional*):
            Number of Telegram Stars that must be paid by non-administrator users of the supergroup chat for each sent message.

        is_paid_messages_available (``bool``, *optional*):
            True, if paid messages are available in this chat.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        display_gifts_button (``bool``, *optional*):
            True, if the gift button should be shown in the message input field for both participants in all chats.
            Returned only in :meth:`~Badmunda.Client.get_chat`

        accepted_gift_types (:obj:`~Badmunda.types.AcceptedGiftTypes`, *optional*):
            Information about gifts that can be received by the user.
            Returned only in :meth:`~Badmunda.Client.get_chat`\

        raw (:obj:`~Badmunda.raw.types.UserFull` | :obj:`~Badmunda.raw.types.ChatFull` | :obj:`~Badmunda.raw.types.ChannelFull`, *optional*):
            The raw chat or user object, as received from the Telegram API.

        full_name (``str``, *property*):
            Full name of the other party in a private chat, for private chats and bots.
    """
    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        id: Optional[int] = None,
        type: Optional["enums.ChatType"] = None,
        is_forum: Optional[bool] = None,
        is_direct_messages_group: Optional[bool] = None,
        is_min: Optional[bool] = None,
        is_members_hidden: Optional[bool] = None,
        is_restricted: Optional[bool] = None,
        is_creator: Optional[bool] = None,
        is_admin: Optional[bool] = None,
        is_deactivated: Optional[bool] = None,
        is_support: Optional[bool] = None,
        is_stories_hidden: Optional[bool] = None,
        is_stories_unavailable: Optional[bool] = None,
        is_business_bot: Optional[bool] = None,
        is_preview: Optional[bool] = None,
        is_banned: Optional[bool] = None,
        is_call_active: Optional[bool] = None,
        is_call_not_empty: Optional[bool] = None,
        is_public: Optional[bool] = None,
        is_paid_reactions_available: Optional[bool] = None,
        verification_status: Optional["types.VerificationStatus"] = None,
        can_send_gift: Optional[bool] = None,
        title: Optional[str] = None,
        username: Optional[str] = None,
        usernames: Optional[List["types.Username"]] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        photo: Optional["types.ChatPhoto"] = None,
        stories: Optional[List["types.Story"]] = None,
        chat_background: Optional["types.ChatBackground"] = None,
        bio: Optional[str] = None,
        description: Optional[str] = None,
        show_message_sender_name: Optional[bool] = None,
        sign_messages: Optional[bool] = None,
        dc_id: Optional[int] = None,
        folder_id: Optional[int] = None,
        has_protected_content: Optional[bool] = None,
        has_visible_history: Optional[bool] = None,
        has_aggressive_anti_spam_enabled: Optional[bool] = None,
        has_automatic_translation: Optional[bool] = None,
        has_forum_tabs: Optional[bool] = None,
        has_direct_messages_group: Optional[bool] = None,
        direct_messages_chat_id: Optional[int] = None,
        invite_link: Optional[str] = None,
        pinned_message: Optional["types.Message"] = None,
        sticker_set_name: Optional[str] = None,
        custom_emoji_sticker_set_name: Optional[str] = None,
        can_set_sticker_set: Optional[bool] = None,
        can_send_paid_media: Optional[bool] = None,
        members: Optional[List["types.User"]] = None,
        members_count: Optional[int] = None,
        restrictions: Optional[List["types.Restriction"]] = None,
        permissions: Optional["types.ChatPermissions"] = None,
        personal_channel: Optional["types.Chat"] = None,
        personal_channel_message: Optional["types.Message"] = None,
        linked_chat: Optional["types.Chat"] = None,
        send_as_chat: Optional["types.Chat"] = None,
        available_reactions: Optional["types.ChatReactions"] = None,
        level: Optional[int] = None,
        reply_color: Optional["types.ChatColor"] = None,
        profile_color: Optional["types.ChatColor"] = None,
        business_away_message: Optional["types.BusinessMessage"] = None,
        business_greeting_message: Optional["types.BusinessMessage"] = None,
        business_work_hours: Optional["types.BusinessMessage"] = None,
        business_location: Optional["types.Location"] = None,
        business_intro: Optional["types.BusinessIntro"] = None,
        birthday: Optional["types.Birthday"] = None,
        message_auto_delete_time: Optional[int] = None,
        unrestrict_boost_count: Optional[int] = None,
        slow_mode_delay: Optional[int] = None,
        slowmode_next_send_date: Optional[datetime] = None,
        join_by_request: Optional[bool] = None,
        join_requests_count: Optional[int] = None,
        banned_until_date: Optional[datetime] = None,
        subscription_until_date: Optional[datetime] = None,
        reactions_limit: Optional[int] = None,
        gift_count: Optional[int] = None,
        bot_verification: Optional["types.BotVerification"] = None,
        settings: Optional["types.ChatSettings"] = None,
        admins_count: Optional[int] = None,
        kicked_count: Optional[int] = None,
        banned_count: Optional[int] = None,
        available_min_id: Optional[int] = None,
        boosts_applied: Optional[int] = None,
        bot_broadcast_admin_rights: Optional["types.ChatPrivileges"] = None,
        bot_group_admin_rights: Optional["types.ChatPrivileges"] = None,
        bot_can_manage_emoji_status: Optional[bool] = None,
        can_delete_channel: Optional[bool] = None,
        can_pin_message: Optional[bool] = None,
        can_schedule_messages: Optional[bool] = None,
        can_set_location: Optional[bool] = None,
        can_set_username: Optional[bool] = None,
        can_view_participants: Optional[bool] = None,
        can_view_revenue: Optional[bool] = None,
        can_view_stars_revenue: Optional[bool] = None,
        can_view_stats: Optional[bool] = None,
        common_chats: Optional[int] = None,
        is_ads_enabled: Optional[bool] = None,
        is_blocked: Optional[bool] = None,
        is_blocked_my_stories_from: Optional[bool] = None,
        is_contact_require_premium: Optional[bool] = None,
        is_phone_calls_available: Optional[bool] = None,
        is_phone_calls_private: Optional[bool] = None,
        is_pinned_stories_available: Optional[bool] = None,
        is_read_dates_available: Optional[bool] = None,
        is_translations_disabled: Optional[bool] = None,
        is_video_calls_available: Optional[bool] = None,
        is_wallpaper_overridden: Optional[bool] = None,
        migrated_from_chat_id: Optional[int] = None,
        migrated_from_max_message_id: Optional[int] = None,
        online_count: Optional[int] = None,
        private_forward_name: Optional[str] = None,
        read_inbox_max_id: Optional[int] = None,
        read_outbox_max_id: Optional[int] = None,
        is_ads_restricted: Optional[bool] = None,
        stats_dc_id: Optional[int] = None,
        theme_emoji: Optional[str] = None,
        unread_count: Optional[int] = None,
        view_forum_as_messages: Optional[bool] = None,
        paid_message_star_count: Optional[int] = None,
        is_paid_messages_available: Optional[bool] = None,
        display_gifts_button: Optional[bool] = None,
        accepted_gift_types: Optional["types.AcceptedGiftTypes"] = None,
        raw: Optional[Union["raw.types.UserFull", "raw.types.ChatFull", "raw.types.ChannelFull"]] = None
    ):
        super().__init__(client)

        self.id = id
        self.type = type
        self.is_forum = is_forum
        self.is_direct_messages_group = is_direct_messages_group
        self.is_min = is_min
        self.is_members_hidden = is_members_hidden
        self.is_restricted = is_restricted
        self.is_creator = is_creator
        self.is_admin = is_admin
        self.is_deactivated = is_deactivated
        self.is_support = is_support
        self.is_stories_hidden = is_stories_hidden
        self.is_stories_unavailable = is_stories_unavailable
        self.is_business_bot = is_business_bot
        self.is_preview = is_preview
        self.is_banned = is_banned
        self.is_call_active = is_call_active
        self.is_call_not_empty = is_call_not_empty
        self.is_public = is_public
        self.is_paid_reactions_available = is_paid_reactions_available
        self.verification_status = verification_status
        self.can_send_gift = can_send_gift
        self.title = title
        self.username = username
        self.usernames = usernames
        self.first_name = first_name
        self.last_name = last_name
        self.photo = photo
        self.stories = stories
        self.chat_background = chat_background
        self.bio = bio
        self.description = description
        self.show_message_sender_name = show_message_sender_name
        self.sign_messages = sign_messages
        self.dc_id = dc_id
        self.folder_id = folder_id
        self.has_protected_content = has_protected_content
        self.has_visible_history = has_visible_history
        self.has_aggressive_anti_spam_enabled = has_aggressive_anti_spam_enabled
        self.has_automatic_translation = has_automatic_translation
        self.has_forum_tabs = has_forum_tabs
        self.has_direct_messages_group = has_direct_messages_group
        self.direct_messages_chat_id = direct_messages_chat_id
        self.invite_link = invite_link
        self.pinned_message = pinned_message
        self.sticker_set_name = sticker_set_name
        self.custom_emoji_sticker_set_name = custom_emoji_sticker_set_name
        self.can_set_sticker_set = can_set_sticker_set
        self.can_send_paid_media = can_send_paid_media
        self.members = members
        self.members_count = members_count
        self.restrictions = restrictions
        self.permissions = permissions
        self.personal_channel = personal_channel
        self.personal_channel_message = personal_channel_message
        self.linked_chat = linked_chat
        self.send_as_chat = send_as_chat
        self.available_reactions = available_reactions
        self.level = level
        self.reply_color = reply_color
        self.profile_color = profile_color
        self.business_away_message = business_away_message
        self.business_greeting_message = business_greeting_message
        self.business_work_hours = business_work_hours
        self.business_location = business_location
        self.business_intro = business_intro
        self.birthday = birthday
        self.message_auto_delete_time = message_auto_delete_time
        self.unrestrict_boost_count = unrestrict_boost_count
        self.slow_mode_delay = slow_mode_delay
        self.slowmode_next_send_date = slowmode_next_send_date
        self.join_by_request = join_by_request
        self.join_requests_count = join_requests_count
        self.banned_until_date = banned_until_date
        self.subscription_until_date = subscription_until_date
        self.reactions_limit = reactions_limit
        self.gift_count = gift_count
        self.bot_verification = bot_verification
        self.settings = settings
        self.admins_count = admins_count
        self.kicked_count = kicked_count
        self.banned_count = banned_count
        self.available_min_id = available_min_id
        self.boosts_applied = boosts_applied
        self.bot_broadcast_admin_rights = bot_broadcast_admin_rights
        self.bot_group_admin_rights = bot_group_admin_rights
        self.bot_can_manage_emoji_status = bot_can_manage_emoji_status
        self.can_delete_channel = can_delete_channel
        self.can_pin_message = can_pin_message
        self.can_schedule_messages = can_schedule_messages
        self.can_set_location = can_set_location
        self.can_set_username = can_set_username
        self.can_view_participants = can_view_participants
        self.can_view_revenue = can_view_revenue
        self.can_view_stars_revenue = can_view_stars_revenue
        self.can_view_stats = can_view_stats
        self.common_chats = common_chats
        self.is_ads_enabled = is_ads_enabled
        self.is_blocked = is_blocked
        self.is_blocked_my_stories_from = is_blocked_my_stories_from
        self.is_contact_require_premium = is_contact_require_premium
        self.is_phone_calls_available = is_phone_calls_available
        self.is_phone_calls_private = is_phone_calls_private
        self.is_pinned_stories_available = is_pinned_stories_available
        self.is_read_dates_available = is_read_dates_available
        self.is_translations_disabled = is_translations_disabled
        self.is_video_calls_available = is_video_calls_available
        self.is_wallpaper_overridden = is_wallpaper_overridden
        self.migrated_from_chat_id = migrated_from_chat_id
        self.migrated_from_max_message_id = migrated_from_max_message_id
        self.online_count = online_count
        self.private_forward_name = private_forward_name
        self.read_inbox_max_id = read_inbox_max_id
        self.read_outbox_max_id = read_outbox_max_id
        self.is_ads_restricted = is_ads_restricted
        self.stats_dc_id = stats_dc_id
        self.theme_emoji = theme_emoji
        self.unread_count = unread_count
        self.view_forum_as_messages = view_forum_as_messages
        self.paid_message_star_count = paid_message_star_count
        self.is_paid_messages_available = is_paid_messages_available
        self.display_gifts_button = display_gifts_button
        self.accepted_gift_types = accepted_gift_types
        self.raw = raw

    # region Deprecated
    # TODO: Remove later

    @property
    def is_verified(self) -> Optional[bool]:
        log.warning(
            "`chat.is_verified` is deprecated and will be removed in future updates. Use `chat.verification_status.is_verified` instead."
        )
        return getattr(self.verification_status, "is_verified", None)

    @property
    def is_scam(self) -> Optional[bool]:
        log.warning(
            "`chat.is_scam` is deprecated and will be removed in future updates. Use `chat.verification_status.is_scam` instead."
        )
        return getattr(self.verification_status, "is_scam", None)

    @property
    def is_fake(self) -> Optional[bool]:
        log.warning(
            "`chat.is_fake` is deprecated and will be removed in future updates. Use `chat.verification_status.is_fake` instead."
        )
        return getattr(self.verification_status, "is_fake", None)

    # endregion

    @staticmethod
    def _parse_user_chat(client, user: "raw.types.User") -> Optional["Chat"]:
        if user is None or isinstance(user, raw.types.UserEmpty):
            return None

        peer_id = user.id

        return Chat(
            id=peer_id,
            type=enums.ChatType.BOT if user.bot else enums.ChatType.PRIVATE,
            is_restricted=user.restricted,
            is_support=user.support,
            is_stories_hidden=user.stories_hidden,
            is_stories_unavailable=user.stories_unavailable,
            is_business_bot=user.bot_business,
            verification_status=types.VerificationStatus._parse(user),
            username=user.username or (user.usernames[0].username if user.usernames else None),
            usernames=types.List([types.Username._parse(r) for r in user.usernames]) or None,
            first_name=user.first_name,
            last_name=user.last_name,
            photo=types.ChatPhoto._parse(client, user.photo, peer_id, user.access_hash),
            restrictions=types.List([types.Restriction._parse(r) for r in user.restriction_reason]) or None,
            dc_id=getattr(getattr(user, "photo", None), "dc_id", None),
            reply_color=types.ChatColor._parse(user.color),
            profile_color=types.ChatColor._parse_profile_color(user.profile_color),
            paid_message_star_count=user.send_paid_messages_stars,
            raw=user,
            client=client
        )

    @staticmethod
    def _parse_chat_chat(client, chat: "raw.types.Chat") -> Optional["Chat"]:
        if chat is None or isinstance(chat, raw.types.ChatEmpty):
            return None

        peer_id = -chat.id
        usernames = getattr(chat, "usernames", [])

        if isinstance(chat, raw.types.ChatForbidden):
            return Chat(
                id=peer_id,
                type=enums.ChatType.GROUP,
                title=chat.title,
                is_banned=True,
                raw=chat,
                client=client
            )

        return Chat(
            id=peer_id,
            type=enums.ChatType.GROUP,
            title=chat.title,
            is_creator=chat.creator,
            is_admin=True if chat.admin_rights else None,
            is_deactivated=chat.deactivated,
            is_call_active=chat.call_active,
            is_call_not_empty=chat.call_not_empty,
            usernames=types.List([types.Username._parse(r) for r in usernames]) or None,
            photo=types.ChatPhoto._parse(client, chat.photo, peer_id, 0),
            permissions=types.ChatPermissions._parse(chat.default_banned_rights),
            members_count=chat.participants_count,
            dc_id=getattr(getattr(chat, "photo", None), "dc_id", None),
            has_protected_content=chat.noforwards,
            raw=chat,
            client=client
        )

    @staticmethod
    def _parse_channel_chat(client, channel: "raw.types.Channel") -> Optional["Chat"]:
        if channel is None:
            return None

        peer_id = utils.get_channel_id(channel.id)
        restriction_reason = getattr(channel, "restriction_reason", [])
        usernames = getattr(channel, "usernames", [])

        if isinstance(channel, raw.types.ChannelForbidden):
            return Chat(
                id=peer_id,
                type=enums.ChatType.SUPERGROUP if channel.megagroup else enums.ChatType.CHANNEL,
                title=channel.title,
                is_banned=True,
                banned_until_date=utils.timestamp_to_datetime(getattr(channel, "until_date", None)),
                raw=channel,
                client=client,
            )

        chat_type = enums.ChatType.CHANNEL

        if channel.monoforum:
            chat_type = enums.ChatType.DIRECT
        elif channel.forum:
            chat_type = enums.ChatType.FORUM
        elif channel.megagroup:
            chat_type = enums.ChatType.SUPERGROUP

        return Chat(
            id=peer_id,
            type=chat_type,
            is_forum=channel.forum,
            is_direct_messages_group=channel.monoforum,
            is_min=channel.min,
            is_restricted=channel.restricted,
            is_creator=channel.creator,
            is_admin=True if channel.admin_rights else None,
            is_stories_hidden=channel.stories_hidden,
            is_stories_unavailable=channel.stories_unavailable,
            is_call_active=channel.call_active,
            is_call_not_empty=channel.call_not_empty,
            verification_status=types.VerificationStatus._parse(channel),
            title=channel.title,
            username=channel.username or (channel.usernames[0].username if channel.usernames else None),
            usernames=types.List([types.Username._parse(r) for r in usernames]) or None,
            photo=types.ChatPhoto._parse(client, channel.photo, peer_id,
                                         getattr(channel, "access_hash", 0)),
            show_message_sender_name=channel.signature_profiles,
            sign_messages=channel.signatures,
            restrictions=types.List([types.Restriction._parse(r) for r in restriction_reason]) or None,
            permissions=types.ChatPermissions._parse(channel.default_banned_rights),
            members_count=channel.participants_count,
            dc_id=getattr(getattr(channel, "photo", None), "dc_id", None),
            has_protected_content=channel.noforwards,
            level=channel.level,
            reply_color=types.ChatColor._parse(channel.color),
            profile_color=types.ChatColor._parse(channel.profile_color),
            subscription_until_date=utils.timestamp_to_datetime(channel.subscription_until_date),
            paid_message_star_count=channel.send_paid_messages_stars,
            has_automatic_translation=channel.autotranslation,
            has_forum_tabs=channel.forum_tabs,
            has_direct_messages_group=channel.broadcast_messages_allowed,
            direct_messages_chat_id=channel.linked_monoforum_id,
            raw=channel,
            client=client
        )

    @staticmethod
    def _parse(
        client,
        message: Union["raw.types.Message", "raw.types.MessageService"],
        users: dict,
        chats: dict,
        is_chat: bool
    ) -> Optional["Chat"]:
        from_id = utils.get_raw_peer_id(message.from_id)
        peer_id = utils.get_raw_peer_id(message.peer_id)
        chat_id = (peer_id or from_id) if is_chat else (from_id or peer_id)

        if isinstance(message.peer_id, raw.types.PeerUser):
            return Chat._parse_user_chat(client, users.get(chat_id))
        elif isinstance(message.peer_id, raw.types.PeerChat):
            return Chat._parse_chat_chat(client, chats.get(chat_id))
        else:
            return Chat._parse_channel_chat(client, chats.get(chat_id))

    @staticmethod
    def _parse_dialog(client, peer, users: dict, chats: dict):
        if isinstance(peer, (raw.types.PeerUser, raw.types.InputPeerUser)):
            return Chat._parse_user_chat(client, users.get(peer.user_id))
        elif isinstance(peer, (raw.types.PeerChat, raw.types.InputPeerChat)):
            return Chat._parse_chat_chat(client, chats.get(peer.chat_id))
        else:
            return Chat._parse_channel_chat(client, chats.get(peer.channel_id))

    @staticmethod
    async def _parse_full_user(client: "Badmunda.Client", user: "raw.types.UserFull", users: dict, chats: dict) -> "Chat":
        parsed_chat = Chat._parse_user_chat(client, users[user.id])
        parsed_chat.raw = user

        parsed_chat.settings = types.ChatSettings._parse(client, user.settings, users)
        # parsed_chat.notify_settings
        parsed_chat.common_chats = user.common_chats_count
        parsed_chat.is_blocked = user.blocked
        parsed_chat.is_phone_calls_available = user.phone_calls_available
        parsed_chat.is_phone_calls_private = user.phone_calls_private
        parsed_chat.can_pin_message = user.can_pin_message
        parsed_chat.can_schedule_messages = user.has_scheduled
        parsed_chat.is_video_calls_available = user.video_calls_available
        parsed_chat.can_send_voice_messages = not user.voice_messages_forbidden
        parsed_chat.is_translations_disabled = user.translations_disabled
        parsed_chat.is_pinned_stories_available = user.stories_pinned_available
        parsed_chat.is_blocked_my_stories_from = user.blocked_my_stories_from
        parsed_chat.is_wallpaper_overridden = user.wallpaper_overridden
        parsed_chat.is_contact_require_premium = user.contact_require_premium
        parsed_chat.is_read_dates_available = not user.read_dates_private
        parsed_chat.is_ads_enabled = user.sponsored_enabled
        parsed_chat.can_view_revenue = user.can_view_revenue
        parsed_chat.bot_can_manage_emoji_status = user.bot_can_manage_emoji_status
        parsed_chat.bio = user.about or None
        # parsed_chat.personal_photo
        # parsed_chat.profile_photo
        # parsed_chat.fallback_photo
        # parsed_chat.bot_info

        if user.pinned_msg_id:
            parsed_chat.pinned_message = await client.get_messages(chat_id=parsed_chat.id, pinned=True)

        parsed_chat.folder_id = user.folder_id
        parsed_chat.message_auto_delete_time = user.ttl_period
        parsed_chat.theme_emoji = user.theme_emoticon
        parsed_chat.private_forward_name = user.private_forward_name
        parsed_chat.bot_group_admin_rights = types.ChatPrivileges._parse(user.bot_group_admin_rights)
        parsed_chat.bot_broadcast_admin_rights = types.ChatPrivileges._parse(user.bot_broadcast_admin_rights)
        # parsed_chat.premium_gifts
        parsed_chat.chat_background = types.ChatBackground._parse(client, user.wallpaper)


        if user.stories:
            parsed_chat.stories = types.List(
                [
                    await types.Story._parse(
                        client, story, user.stories.peer, users, chats
                    )
                    for story in user.stories.stories
                ]
            ) or None

        parsed_chat.business_work_hours = types.BusinessWorkingHours._parse(user.business_work_hours)
        parsed_chat.business_location = types.Location._parse(client, user.business_location)
        parsed_chat.business_greeting_message = types.BusinessMessage._parse(client, user.business_greeting_message, users)
        parsed_chat.business_away_message = types.BusinessMessage._parse(client, user.business_away_message, users)
        parsed_chat.business_intro = await types.BusinessIntro._parse(client, user.business_intro)
        parsed_chat.birthday = types.Birthday._parse(user.birthday)

        if user.personal_channel_id:
            parsed_chat.personal_channel = Chat._parse_channel_chat(client, chats[user.personal_channel_id])
            parsed_chat.personal_channel_message = await client.get_messages(
                chat_id=parsed_chat.personal_channel.id,
                message_ids=user.personal_channel_message
            )

        parsed_chat.gift_count = user.stargifts_count
        # parsed_chat.starref_program
        parsed_chat.bot_verification = types.BotVerification._parse(
            client,
            user.bot_verification,
            users
        )
        parsed_chat.paid_message_star_count = user.send_paid_messages_stars
        parsed_chat.display_gifts_button = user.display_gifts_button
        parsed_chat.accepted_gift_types = types.AcceptedGiftTypes._parse(user.disallowed_gifts)

        return parsed_chat

    @staticmethod
    async def _parse_full_chat(client: "Badmunda.Client", chat: "raw.types.ChatFull", users: dict, chats: dict) -> "Chat":
        parsed_chat = Chat._parse_chat_chat(client, chats[chat.id])
        parsed_chat.raw = chat

        parsed_chat.description = chat.about or None

        if isinstance(chat.participants, raw.types.ChatParticipants):
            parsed_chat.members_count = len(chat.participants.participants)

        # parsed_chat.notify_settings
        parsed_chat.can_set_username = chat.can_set_username
        parsed_chat.can_schedule_messages = chat.has_scheduled
        parsed_chat.is_translations_disabled = chat.translations_disabled

        if isinstance(chat.exported_invite, raw.types.ChatInviteExported):
            parsed_chat.invite_link = chat.exported_invite.link

        # parsed_chat.bot_info

        if chat.pinned_msg_id:
            parsed_chat.pinned_message = await client.get_messages(chat_id=parsed_chat.id, pinned=True)

        parsed_chat.folder_id = chat.folder_id
        # parsed_chat.call
        parsed_chat.message_auto_delete_time = chat.ttl_period
        # parsed_chat.groupcall_default_join_as
        parsed_chat.theme_emoji = chat.theme_emoticon
        parsed_chat.join_requests_count = chat.requests_pending
        # parsed_chat.recent_requesters
        parsed_chat.available_reactions = types.ChatReactions._parse(client, chat.available_reactions)
        parsed_chat.reactions_limit = chat.reactions_limit

        return parsed_chat

    @staticmethod
    async def _parse_full_channel(client: "Badmunda.Client", channel: "raw.types.ChannelFull", users, chats) -> "Chat":
        parsed_chat = Chat._parse_channel_chat(client, chats[channel.id])
        parsed_chat.raw = channel

        parsed_chat.description = channel.about or None
        parsed_chat.read_inbox_max_id = channel.read_inbox_max_id
        parsed_chat.read_outbox_max_id = channel.read_outbox_max_id
        parsed_chat.unread_count = channel.unread_count
        # parsed_chat.chat_photo
        # parsed_chat.notify_settings
        # parsed_chat.bot_info
        # parsed_chat.pts
        parsed_chat.can_view_participants = channel.can_view_participants
        parsed_chat.can_set_username = channel.can_set_username
        parsed_chat.can_set_sticker_set = channel.can_set_stickers
        parsed_chat.has_visible_history = channel.hidden_prehistory
        parsed_chat.can_set_location = channel.can_set_location
        parsed_chat.can_schedule_messages = channel.has_scheduled
        parsed_chat.can_view_stats = channel.can_view_stats
        parsed_chat.is_blocked = channel.blocked
        parsed_chat.can_delete_channel = channel.can_delete_channel
        parsed_chat.has_aggressive_anti_spam_enabled = channel.antispam
        parsed_chat.is_members_hidden = channel.participants_hidden
        parsed_chat.is_translations_disabled = channel.translations_disabled
        parsed_chat.is_pinned_stories_available = channel.stories_pinned_available
        parsed_chat.view_forum_as_messages = channel.view_forum_as_messages
        parsed_chat.is_ads_restricted = channel.restricted_sponsored
        parsed_chat.can_view_revenue = channel.can_view_revenue
        parsed_chat.can_send_paid_media = channel.paid_media_allowed
        parsed_chat.can_view_stars_revenue = channel.can_view_stars_revenue
        parsed_chat.is_paid_reactions_available = channel.paid_messages_available
        parsed_chat.can_send_gift = channel.stargifts_available
        parsed_chat.members_count = channel.participants_count
        parsed_chat.admins_count = channel.admins_count
        parsed_chat.kicked_count = channel.kicked_count
        parsed_chat.banned_count = channel.banned_count
        parsed_chat.online_count = channel.online_count

        if isinstance(channel.exported_invite, raw.types.ChatInviteExported):
            parsed_chat.invite_link = channel.exported_invite.link

        parsed_chat.migrated_from_chat_id = channel.migrated_from_chat_id
        parsed_chat.migrated_from_max_id = channel.migrated_from_max_id

        if channel.pinned_msg_id:
            parsed_chat.pinned_message = await client.get_messages(chat_id=parsed_chat.id, pinned=True)

        # parsed_chat.stickerset
        parsed_chat.available_min_id = channel.available_min_id
        parsed_chat.folder_id = channel.folder_id

        if chats.get(channel.linked_chat_id):
            parsed_chat.linked_chat = Chat._parse_channel_chat(client, chats[channel.linked_chat_id])

        # parsed_chat.location
        parsed_chat.slow_mode_delay = channel.slowmode_seconds
        parsed_chat.slowmode_next_send_date = utils.timestamp_to_datetime(
            channel.slowmode_next_send_date
        )
        parsed_chat.stats_dc_id = channel.stats_dc
        # parsed_chat.call
        parsed_chat.message_auto_delete_time = channel.ttl_period
        # parsed_chat.pending_suggestions
        # parsed_chat.groupcall_default_join_as
        parsed_chat.theme_emoji = channel.theme_emoticon
        parsed_chat.join_requests_count = channel.requests_pending
        # parsed_chat.recent_requesters

        if channel.default_send_as:
            if isinstance(channel.default_send_as, raw.types.PeerUser):
                send_as_raw = users[channel.default_send_as.user_id]
            else:
                send_as_raw = chats[channel.default_send_as.channel_id]

            parsed_chat.send_as_chat = Chat._parse_chat(client, send_as_raw)

        parsed_chat.available_reactions = types.ChatReactions._parse(client, channel.available_reactions)
        parsed_chat.reactions_limit = channel.reactions_limit

        if channel.stories:
            parsed_chat.stories = types.List(
                [
                    await types.Story._parse(
                        client, story, channel.stories.peer, users, chats
                    )
                    for story in channel.stories.stories
                ]
            ) or None

        parsed_chat.chat_background = types.ChatBackground._parse(client, channel.wallpaper)
        parsed_chat.boosts_applied = channel.boosts_applied
        parsed_chat.unrestrict_boost_count = channel.boosts_unrestrict
        parsed_chat.custom_emoji_sticker_set_name = getattr(channel.emojiset, "short_name", None)
        parsed_chat.bot_verification = types.BotVerification._parse(
            client,
            channel.bot_verification,
            users
        )
        parsed_chat.gift_count = channel.stargifts_count
        parsed_chat.sticker_set_name = getattr(channel.stickerset, "short_name", None)
        parsed_chat.is_paid_messages_available = channel.paid_messages_available

        return parsed_chat

    @staticmethod
    async def _parse_full(client: "Badmunda.Client", chat_full: Union["raw.types.UserFull", "raw.types.ChatFull", "raw.types.ChannelFull"]) -> Optional["Chat"]:
        users = {u.id: u for u in chat_full.users}
        chats = {c.id: c for c in chat_full.chats}

        if isinstance(chat_full, raw.types.users.UserFull):
            return await Chat._parse_full_user(client, chat_full.full_user, users, chats)
        elif isinstance(chat_full, raw.types.messages.ChatFull) and isinstance(chat_full.full_chat, raw.types.ChatFull):
            return await Chat._parse_full_chat(client, chat_full.full_chat, users, chats)
        elif isinstance(chat_full, raw.types.messages.ChatFull) and isinstance(chat_full.full_chat, raw.types.ChannelFull):
            return await Chat._parse_full_channel(client, chat_full.full_chat, users, chats)

    @staticmethod
    def _parse_chat(client, chat: Union[raw.types.Chat, raw.types.User, raw.types.Channel]) -> Optional["Chat"]:
        if isinstance(chat, (raw.types.Chat, raw.types.ChatForbidden)):
            return Chat._parse_chat_chat(client, chat)
        elif isinstance(chat, raw.types.User):
            return Chat._parse_user_chat(client, chat)
        else:
            return Chat._parse_channel_chat(client, chat)

    @staticmethod
    def _parse_preview(client, chat_invite: "raw.types.ChatInvite") -> "Chat":
        return Chat(
            type=(
                enums.ChatType.SUPERGROUP if chat_invite.megagroup else
                enums.ChatType.CHANNEL if chat_invite.broadcast else
                enums.ChatType.GROUP
            ),
            is_public=chat_invite.public,
            is_preview=True,
            verification_status=types.VerificationStatus._parse(chat_invite),
            title=chat_invite.title,
            photo=types.Photo._parse(client, chat_invite.photo),
            members_count=chat_invite.participants_count,
            members=[
                types.User._parse(client, user)
                for user in getattr(chat_invite, "participants", [])
            ] or None,
            description=chat_invite.about or None,
            join_by_request=chat_invite.request_needed,
            profile_color=types.ChatColor._parse(chat_invite.color),
            raw=chat_invite,
            client=client
        )

    @property
    def full_name(self) -> str:
        return " ".join(filter(None, [self.first_name, self.last_name])) or self.title or None

    async def archive(self):
        """Bound method *archive* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.archive_chats(-100123456789)

        Example:
            .. code-block:: python

                await chat.archive()

        Returns:
            True on success.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.archive_chats(self.id)

    async def unarchive(self):
        """Bound method *unarchive* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.unarchive_chats(-100123456789)

        Example:
            .. code-block:: python

                await chat.unarchive()

        Returns:
            True on success.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.unarchive_chats(self.id)

    # TODO: Remove notes about "All Members Are Admins" for basic groups, the attribute doesn't exist anymore
    async def set_title(self, title: str) -> bool:
        """Bound method *set_title* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.set_chat_title(
                chat_id=chat_id,
                title=title
            )

        Example:
            .. code-block:: python

                await chat.set_title("Lounge")

        .. note::

            In regular groups (non-supergroups), this method will only work if the "All Members Are Admins"
            setting is off.

        Parameters:
            title (``str``):
                New chat title, 1-255 characters.

        Returns:
            ``bool``: True on success.

        Raises:
            RPCError: In case of Telegram RPC error.
            ValueError: In case a chat_id belongs to user.
        """
        return await self._client.set_chat_title(
            chat_id=self.id,
            title=title
        )

    async def set_description(self, description: str) -> bool:
        """Bound method *set_description* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.set_chat_description(
                chat_id=chat_id,
                description=description
            )

        Example:
            .. code-block:: python

                await chat.set_chat_description("Don't spam!")

        Parameters:
            description (``str``):
                New chat description, 0-255 characters.

        Returns:
            ``bool``: True on success.

        Raises:
            RPCError: In case of Telegram RPC error.
            ValueError: If a chat_id doesn't belong to a supergroup or a channel.
        """
        return await self._client.set_chat_description(
            chat_id=self.id,
            description=description
        )

    async def set_photo(
        self,
        *,
        photo: Union[str, BinaryIO] = None,
        video: Union[str, BinaryIO] = None,
        video_start_ts: float = None,
    ) -> bool:
        """Bound method *set_photo* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.set_chat_photo(
                chat_id=chat_id,
                photo=photo
            )

        Example:
            .. code-block:: python

                # Set chat photo using a local file
                await chat.set_photo(photo="photo.jpg")

                # Set chat photo using an existing Photo file_id
                await chat.set_photo(photo=photo.file_id)


                # Set chat video using a local file
                await chat.set_photo(video="video.mp4")

                # Set chat photo using an existing Video file_id
                await chat.set_photo(video=video.file_id)

        Parameters:
            photo (``str`` | ``BinaryIO``, *optional*):
                New chat photo. You can pass a :obj:`~Badmunda.types.Photo` file_id, a file path to upload a new photo
                from your local machine or a binary file-like object with its attribute
                ".name" set for in-memory uploads.

            video (``str`` | ``BinaryIO``, *optional*):
                New chat video. You can pass a :obj:`~Badmunda.types.Video` file_id, a file path to upload a new video
                from your local machine or a binary file-like object with its attribute
                ".name" set for in-memory uploads.

            video_start_ts (``float``, *optional*):
                The timestamp in seconds of the video frame to use as photo profile preview.

        Returns:
            ``bool``: True on success.

        Raises:
            RPCError: In case of a Telegram RPC error.
            ValueError: if a chat_id belongs to user.
        """
        return await self._client.set_chat_photo(
            chat_id=self.id,
            photo=photo,
            video=video,
            video_start_ts=video_start_ts
        )

    async def set_ttl(self, ttl_seconds: int) -> "types.Message":
        """Bound method *set_ttl* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.set_chat_ttl(
                chat_id=chat_id,
                ttl_seconds=ttl_seconds
            )

        Example:
            .. code-block:: python

                await chat.set_ttl(86400)

        Returns:
            :obj:`~Badmunda.types.Message`: On success, the generated service message is returned.
        """
        return await self._client.set_chat_ttl(
            chat_id=self.id,
            ttl_seconds=ttl_seconds
        )

    async def ban_member(
        self,
        user_id: Union[int, str],
        until_date: datetime = utils.zero_datetime()
    ) -> Union["types.Message", bool]:
        """Bound method *ban_member* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.ban_chat_member(
                chat_id=chat_id,
                user_id=user_id
            )

        Example:
            .. code-block:: python

                await chat.ban_member(123456789)

        .. note::

            In regular groups (non-supergroups), this method will only work if the "All Members Are Admins" setting is
            off in the target group. Otherwise members may only be removed by the group's creator or by the member
            that added them.

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target user.
                For a contact that exists in your Telegram address book you can use his phone number (str).

            until_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the user will be unbanned.
                If user is banned for more than 366 days or less than 30 seconds from the current time they are
                considered to be banned forever. Defaults to epoch (ban forever).

        Returns:
            :obj:`~Badmunda.types.Message` | ``bool``: On success, a service message will be returned (when applicable), otherwise, in
            case a message object couldn't be returned, True is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.ban_chat_member(
            chat_id=self.id,
            user_id=user_id,
            until_date=until_date
        )

    async def unban_member(
        self,
        user_id: Union[int, str]
    ) -> bool:
        """Bound method *unban_member* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.unban_chat_member(
                chat_id=chat_id,
                user_id=user_id
            )

        Example:
            .. code-block:: python

                await chat.unban_member(123456789)

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target user.
                For a contact that exists in your Telegram address book you can use his phone number (str).

        Returns:
            ``bool``: True on success.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.unban_chat_member(
            chat_id=self.id,
            user_id=user_id,
        )

    async def restrict_member(
        self,
        user_id: Union[int, str],
        permissions: "types.ChatPermissions",
        until_date: datetime = utils.zero_datetime(),
    ) -> "types.Chat":
        """Bound method *unban_member* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.restrict_chat_member(
                chat_id=chat_id,
                user_id=user_id,
                permissions=ChatPermissions()
            )

        Example:
            .. code-block:: python

                await chat.restrict_member(user_id, ChatPermissions())

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target user.
                For a contact that exists in your Telegram address book you can use his phone number (str).

            permissions (:obj:`~Badmunda.types.ChatPermissions`):
                New user permissions.

            until_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the user will be unbanned.
                If user is banned for more than 366 days or less than 30 seconds from the current time they are
                considered to be banned forever. Defaults to epoch (ban forever).

        Returns:
            :obj:`~Badmunda.types.Chat`: On success, a chat object is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.restrict_chat_member(
            chat_id=self.id,
            user_id=user_id,
            permissions=permissions,
            until_date=until_date,
        )

    # Set None as privileges default due to issues with partially initialized module, because at the time Chat
    # is being initialized, ChatPrivileges would be required here, but was not initialized yet.
    async def promote_member(
        self,
        user_id: Union[int, str],
        privileges: "types.ChatPrivileges" = None
    ) -> bool:
        """Bound method *promote_member* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.promote_chat_member(
                chat_id=chat_id,
                user_id=user_id
            )

        Example:

            .. code-block:: python

                await chat.promote_member(123456789)

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target user.
                For a contact that exists in your Telegram address book you can use his phone number (str).

            privileges (:obj:`~Badmunda.types.ChatPrivileges`, *optional*):
                New user privileges.

        Returns:
            ``bool``: True on success.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.promote_chat_member(
            chat_id=self.id,
            user_id=user_id,
            privileges=privileges
        )

    async def join(self):
        """Bound method *join* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.join_chat(123456789)

        Example:
            .. code-block:: python

                await chat.join()

        .. note::

            This only works for public groups, channels that have set a username or linked chats.

        Returns:
            :obj:`~Badmunda.types.Chat`: On success, a chat object is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.join_chat(self.username or self.id)

    async def leave(self):
        """Bound method *leave* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.leave_chat(123456789)

        Example:
            .. code-block:: python

                await chat.leave()

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self._client.leave_chat(self.id)

    async def export_invite_link(self):
        """Bound method *export_invite_link* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            client.export_chat_invite_link(123456789)

        Example:
            .. code-block:: python

                chat.export_invite_link()

        Returns:
            ``str``: On success, the exported invite link is returned.

        Raises:
            ValueError: In case the chat_id belongs to a user.
        """
        return await self._client.export_chat_invite_link(self.id)

    async def get_member(
        self,
        user_id: Union[int, str],
    ) -> "types.ChatMember":
        """Bound method *get_member* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.get_chat_member(
                chat_id=chat_id,
                user_id=user_id
            )

        Example:
            .. code-block:: python

                await chat.get_member(user_id)

        Returns:
            :obj:`~Badmunda.types.ChatMember`: On success, a chat member is returned.
        """
        return await self._client.get_chat_member(
            self.id,
            user_id=user_id
        )

    def get_members(
        self,
        query: str = "",
        limit: int = 0,
        filter: "enums.ChatMembersFilter" = enums.ChatMembersFilter.SEARCH
    ) -> AsyncGenerator["types.ChatMember", None]:
        """Bound method *get_members* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            async for member in client.get_chat_members(chat_id):
                print(member)

        Example:
            .. code-block:: python

                async for member in chat.get_members():
                    print(member)

        Parameters:
            query (``str``, *optional*):
                Query string to filter members based on their display names and usernames.
                Only applicable to supergroups and channels. Defaults to "" (empty string).
                A query string is applicable only for :obj:`~Badmunda.enums.ChatMembersFilter.SEARCH`,
                :obj:`~Badmunda.enums.ChatMembersFilter.BANNED` and :obj:`~Badmunda.enums.ChatMembersFilter.RESTRICTED`
                filters only.

            limit (``int``, *optional*):
                Limits the number of members to be retrieved.

            filter (:obj:`~Badmunda.enums.ChatMembersFilter`, *optional*):
                Filter used to select the kind of members you want to retrieve. Only applicable for supergroups
                and channels.

        Returns:
            ``Generator``: On success, a generator yielding :obj:`~Badmunda.types.ChatMember` objects is returned.
        """
        return self._client.get_chat_members(
            self.id,
            query=query,
            limit=limit,
            filter=filter
        )

    async def add_members(
        self,
        user_ids: Union[Union[int, str], List[Union[int, str]]],
        forward_limit: int = 100
    ) -> bool:
        """Bound method *add_members* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.add_chat_members(chat_id, user_id)

        Example:
            .. code-block:: python

                await chat.add_members(user_id)

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.add_chat_members(
            self.id,
            user_ids=user_ids,
            forward_limit=forward_limit
        )

    async def mark_unread(self, ) -> bool:
        """Bound method *mark_unread* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.mark_unread(chat_id)

        Example:
            .. code-block:: python

                await chat.mark_unread()

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.mark_chat_unread(self.id)

    async def set_protected_content(self, enabled: bool) -> bool:
        """Bound method *set_protected_content* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            await client.set_chat_protected_content(chat_id, enabled)

        Parameters:
            enabled (``bool``):
                Pass True to enable the protected content setting, False to disable.

        Example:
            .. code-block:: python

                await chat.set_protected_content(enabled)

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.set_chat_protected_content(
            self.id,
            enabled=enabled
        )

    async def unpin_all_messages(self) -> bool:
        """Bound method *unpin_all_messages* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            client.unpin_all_chat_messages(chat_id)

        Example:
            .. code-block:: python

                chat.unpin_all_messages()

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.unpin_all_chat_messages(self.id)

    async def mute(self, mute_until: datetime = None) -> bool:
        """Bound method *mute* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            client.update_chat_notifications(chat_id, mute=True, mute_until=mute_until)

        Parameters:
            mute (``bool``, *optional*):
                Pass True if you want to mute chat.

            until_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the user will be unmuted. Defaults to forever.

        Example:
            .. code-block:: python

                chat.mute()

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.update_chat_notifications(self.id, mute=True, mute_until=mute_until)

    async def unmute(self) -> bool:
        """Bound method *unmute* of :obj:`~Badmunda.types.Chat`.

        Use as a shortcut for:

        .. code-block:: python

            client.update_chat_notifications(chat_id, mute=False)

        Example:
            .. code-block:: python

                chat.unmute()

        Returns:
            ``bool``: On success, True is returned.
        """
        return await self._client.update_chat_notifications(self.id, mute=False)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

import Badmunda
from Badmunda import raw

from ..object import Object


class RestrictionReason(Object):
    """Restriction reason.

    Parameters:
        platform (``str``):
            Platform identifier (ios, android, wp, all, etc.), can be concatenated with a dash as separator (android-ios, ios-wp, etc).

        reason (``str``):
            Restriction reason (porno, terms, etc.). Ignore this restriction reason if it is contained in the ignore_restriction_reasons » client configuration parameter.

        text (``str``):
            Error message to be shown to the user.
    """
    def __init__(
        self, *,
        platform: str,
        reason: str,
        text: str
    ):
        super().__init__()

        self.platform = platform
        self.reason = reason
        self.text = text

    @staticmethod
    def _parse(
        restriction_reason: "raw.types.RestrictionReason"
    ) -> Optional["RestrictionReason"]:
        if not restriction_reason:
            return None

        return RestrictionReason(
            platform=restriction_reason.platform,
            reason=restriction_reason.reason,
            text=restriction_reason.text
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import AsyncGenerator, Optional

import Badmunda
from Badmunda import raw
from Badmunda import types


class GetAllStories:
    async def get_all_stories(
        self: "Badmunda.Client",
        next: Optional[bool] = None,
        hidden: Optional[bool] = None,
        state: Optional[str] = None,
    ) -> AsyncGenerator["types.Story", None]:
        """Get all active or hidden stories that displayed on the action bar on the homescreen.

        .. include:: /_includes/usable-by/users.rst

        Parameters
            next (``bool``, *optional*):
                If next and state are both set, uses the passed state to paginate to the next results.
                If neither state nor next are set, fetches the initial page.
                If state is set and next is not set, check for changes in the active/hidden peerset.

            hidden (``bool``, *optional*):
                If set, fetches the hidden active story list, otherwise fetches the active story list.

            state (``str``, *optional*):
                If next and state are both set, uses the passed state to paginate to the next results.
                If neither state nor next are set, fetches the initial page.
                If state is set and next is not set, check for changes in the active/hidden peerset.

        Returns:
            ``Generator``: On success, a generator yielding :obj:`~Badmunda.types.Story` objects is returned.

        Example:
            .. code-block:: python

                # Get all active story
                async for story in app.get_all_stories():
                    print(story)
        """

        r = await self.invoke(
            raw.functions.stories.GetAllStories(
                next=next,
                hidden=hidden,
                state=state
            )
        )

        users = {i.id: i for i in r.users}
        chats = {i.id: i for i in r.chats}

        for peer_story in r.peer_stories:
            for story in peer_story.stories:
                yield await types.Story._parse(
                    self,
                    story,
                    peer_story.peer,
                    users,
                    chats
                )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import Badmunda
from Badmunda import raw


class TransferBusinessAccountStars:
    async def transfer_business_account_stars(
        self: "Badmunda.Client",
        business_connection_id: str,
        star_count: int,
    ) -> bool:
        """Transfers Telegram Stars from the business account balance to the bot’s balance.

        .. note::

            Requires the `can_transfer_stars` business bot right.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            business_connection_id (``str``):
                Unique identifier of the business connection.

            star_count (``int`` | ``str``):
                Number of Telegram Stars to transfer, 1-10000.

        Returns:
            ``bool``: On success, True is returned.
        """
        # Why telegram won't let us just use InputPeerSelf :(
        if self.me:
            bot_id = self.me.id
        else:
            bot_id = (
                await self.invoke(raw.functions.users.GetUsers(id=[raw.types.InputPeerSelf()]))
            )[0].id

        invoice = raw.types.InputInvoiceBusinessBotTransferStars(
            bot=await self.resolve_peer(bot_id), stars=star_count
        )

        payment_form = await self.invoke(
            raw.functions.payments.GetPaymentForm(invoice=invoice),
            business_connection_id=business_connection_id,
        )

        await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=payment_form.form_id,
                invoice=invoice,
            ),
            business_connection_id=business_connection_id,
        )

        return True



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import Badmunda
from Badmunda import raw


class UpdateStatus:
    async def update_status(
        self: "Badmunda.Client",
        offline: bool = False
    ) -> bool:
        """Update your profile status.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            offline (``bool``):
                The new status. Pass True to appear offline.

        Returns:
            ``bool``: True on success.

        Example:
            .. code-block:: python

                # Change status to online
                await app.update_status()

                # Change status to offline
                await app.update_status(offline=True)
        """
        r = await self.invoke(
            raw.functions.account.UpdateStatus(
                offline=offline
            )
        )

        return bool(r)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Dict, List, Optional

import Badmunda
from Badmunda import raw, types, utils

from ..object import Object


class Checklist(Object):
    """Describes a checklist.

    Parameters:
        title (``str``):
            Title of the checklist.

        entities (List of :obj:`~Badmunda.types.MessageEntity`, *optional*):
            Entities in the title of the checklist.
            May contain only Bold, Italic, Underline, Strikethrough, Spoiler, and CustomEmoji entities.

        tasks (List of :obj:`~Badmunda.types.ChecklistTask`, *optional*):
            List of tasks in the checklist.

        others_can_add_tasks (``bool``, *optional*):
            True, if users other than creator of the list can add tasks to the list.

        can_add_tasks (``bool``, *optional*):
            True, if the current user can add tasks to the list if they have Telegram Premium subscription.

        others_can_mark_tasks_as_done (``bool``, *optional*):
            True, if users other than creator of the list can mark tasks as done or not done.
            If True, then the checklist is called "group checklist".

        can_mark_tasks_as_done (``bool``, *optional*):
            True, if the current user can mark tasks as done or not done if they have Telegram Premium subscription.
    """

    def __init__(
        self,
        *,
        title: str,
        entities: Optional[List["types.MessageEntity"]] = None,
        tasks: Optional[List["types.ChecklistTask"]] = None,
        others_can_add_tasks: Optional[bool] = None,
        can_add_tasks: Optional[bool] = None,
        others_can_mark_tasks_as_done: Optional[bool] = None,
        can_mark_tasks_as_done: Optional[bool] = None,
    ):
        super().__init__()

        self.title = title
        self.entities = entities
        self.tasks = tasks
        self.others_can_add_tasks = others_can_add_tasks
        self.can_add_tasks = can_add_tasks
        self.others_can_mark_tasks_as_done = others_can_mark_tasks_as_done
        self.can_mark_tasks_as_done = can_mark_tasks_as_done

    @staticmethod
    def _parse(
        client: "Badmunda.Client",
        checklist: "raw.types.MessageMediaToDo",
        users: Dict[int, "raw.base.User"],
    ) -> "Checklist":
        completions = {i.id: i for i in getattr(checklist, "completions", [])}

        checklist_tasks = []

        for task in checklist.todo.list:
            checklist_tasks.append(
                types.ChecklistTask._parse(
                    client,
                    task,
                    completions.get(task.id),
                    users
                )
            )

        title, entities = (
            utils.parse_text_with_entities(client, checklist.todo.title, users)
        ).values()

        return Checklist(
            title=title,
            entities=entities,
            tasks=checklist_tasks,
            others_can_add_tasks=checklist.todo.others_can_append,
            # can_add_tasks=checklist.todo.can_append,
            others_can_mark_tasks_as_done=checklist.todo.others_can_complete,
            # can_mark_tasks_as_done=checklist.todo.can_complete,
        )
